//
// iflcopy:
//
//    A simple image copying program using the Image File Library (IFL)
//

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#include <ifl/iflError.h>
#include <ifl/iflFormat.h>
#include <ifl/iflFile.h>
#include <ifl/iflDataSize.h>
#include <ifl/iflConfig.h>
#include <ifl/iflTileIter.h>
#include <ifl/iflMinMax.h>

// simple routine to decode IFL status, print message and exit

void
errorExit(char* prefix, iflStatus err)
{
    char msg[1024];
    iflStatusToString(err, msg, sizeof msg);
    fprintf(stderr, "iflcopy: %s: %s\n", prefix, msg);
    exit(EXIT_FAILURE);
}


void
main(int argc, char* argv[])
{
    if (argc != 3) {
	fprintf(stderr, "Usage: iflcopy <infilename> <outfilename>\n");
	exit(EXIT_FAILURE);
    }

    iflStatus err;
    iflFile *in = iflFile::open(argv[1], O_RDONLY, &err);
    if (in == NULL) errorExit("Unable to open input file", err);

    iflFile *out = iflFile::create(argv[2], in, NULL, NULL, &err);
    if (out == NULL) errorExit("Unable to create output file", err);

    iflSize pgSize;
    out->getPageSize(pgSize, out->getOrientation());
    int bufsize = iflDataSize(in->getDataType(), pgSize);
    void* buf = new char [bufsize];
    if (buf == NULL) {
	fprintf(stderr, "iflcopy: unable to allocate %u bytes\n", bufsize);
	exit(EXIT_FAILURE);
    }
    if (in->getColorModel() == iflRGBPalette) {
    	const iflColormap* cmap;
    	in->getColormap(cmap);
	out->setColormap(cmap);
    }

    iflSize size;
    out->getSize(size, out->getOrientation());
    iflConfig config(out->getDataType(), out->getOrder(), size.c,
		     NULL, 0, out->getOrientation());
    iflTileIter iter(iflTile3Dint(0, 0, 0, size.x, size.y, size.z),
		     pgSize, size.c);
    while (iter.more()) {
	iflSize rwSize(iflMin(size.x - iter.x, pgSize.x),
		       iflMin(size.y - iter.y, pgSize.y),
		       iflMin(size.z - iter.z, pgSize.z),
		       iflMin(size.c - iter.c, pgSize.c));

	err = in->getTile(iter.x, iter.y, iter.z,
			  rwSize.x, rwSize.y, rwSize.z,
			  buf, &config);
	if (err != iflOKAY) errorExit("Couldn't read page from input", err);

	err = out->setPage(buf, iter.x, iter.y, iter.z, iter.c,
			   rwSize.x, rwSize.y, rwSize.z, rwSize.c);
	if (err != iflOKAY) errorExit("Couldn't write page to output", err);
    }

    delete[] buf;
    err = out->flush();
    if (err != iflOKAY) errorExit("Error flushing output file", err);
    err = out->close();
    if (err != iflOKAY) errorExit("Error closing output file", err);
    err = in->close();
    if (err != iflOKAY) errorExit("Error closing input file", err);
}
