#include <ifl/iflFile.h>
#include <ifl/iflFormat.h>
#include <ifl/iflError.h>
#include <ifl/iflDataSize.h>
#include <ifl/iflColor.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#ifdef _WIN32
#include <io.h>
typedef __int64 off64_t;
#define oserror() errno
#define ssize_t int
#define lseek64 _lseeki64
#else
#include <unistd.h>
#endif

#if defined(__sgi)
#define RAW_BUILD_DIO
#define RAW_DIO_MIN_PAGESIZE	size_t(65536)
#endif

#if defined(__sgi) && IFL_HAVE_IRIX_RELEASE(6,5)
#define PREAD pread64
#define PWRITE pwrite64
#else
#define PREAD pread
#define PWRITE pwrite
#endif

enum { 
    SizeParam, PageSizeParam, ColorModelParam, DataTypeParam, OrderParam, 
    OrientationParam, MinimumParam, MaximumParam, OffsetParam 
};

static iflNameDesc paramNames[] = {
    { "size", SizeParam }, 
    { "pagesize", PageSizeParam }, 
    { "colormodel", ColorModelParam }, 
    { "type", DataTypeParam }, 
    { "order", OrderParam }, 
    { "orientation", OrientationParam }, 
    { "minimum", MinimumParam }, 
    { "maximum", MaximumParam }, 
    { "offset", OffsetParam }, 
    { NULL, -1 }
};

static iflNameDesc colorModelNames[] = {
    { "luminance", iflLuminance }, 
    { "luminancealpha", iflLuminanceAlpha }, 
    { "la", iflLuminanceAlpha }, 
    { "negative", iflNegative }, 
    { "palette", iflRGBPalette }, 
    { "rgb", iflRGB }, 
    { "rgba", iflRGBA }, 
    { "hsv", iflHSV }, 
    { "cmy", iflCMY }, 
    { "cmyk", iflCMYK }, 
    { "bgr", iflBGR }, 
    { "abgr", iflABGR }, 
    { "ycc", iflYCC }, 
    { "multispectral", iflMultiSpectral }, 
    { NULL, -1 }
};

static iflNameDesc dataTypeNames[] = {
    { "bit", iflBit }, 
    { "char", iflChar }, 
    { "uchar", iflUChar }, 
    { "short", iflShort }, 
    { "ushort", iflUShort }, 
    { "int", iflInt }, 
    { "uint", iflUInt }, 
    { "float", iflFloat }, 
    { "double", iflDouble }, 
    { NULL, -1 }
};

static iflNameDesc orderNames[] = {
    { "interleaved", iflInterleaved }, 
    { "separate", iflSeparate }, 
    { "sequential", iflSequential }, 
    { NULL, -1 }
};

static iflNameDesc orientationNames[] = {
    { "tl", iflUpperLeftOrigin }, 
    { "tr", iflUpperRightOrigin }, 
    { "bl", iflLowerLeftOrigin }, 
    { "br", iflLowerRightOrigin }, 
    { "lt", iflLeftUpperOrigin }, 
    { "rt", iflRightUpperOrigin }, 
    { "rb", iflRightLowerOrigin }, 
    { "lb", iflLeftLowerOrigin }, 
    { NULL, -1 }
};


class iflRawFile : public iflFile {

public:
    virtual iflStatus getScaleMinMax(double& min, double& max);
    virtual iflStatus setScaleMinMax(double min, double max);

    virtual iflStatus getPage(void* data, int,int,int,int, int,int,int,int);
    virtual iflStatus setPage(const void* data, int,int,int,int,
			      int,int,int,int);

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    virtual iflStatus _enableDirectIO(int enable);
    virtual iflStatus _getDirectIOMemAlignment(size_t& memAlign);
#endif

private:
    virtual iflStatus openFile();
    virtual iflStatus createFile();
    virtual iflStatus closeFile();
    friend class iflRawFormat;
    iflRawFile() {}  // only called by iflRawFormat::newfileobj()
    ~iflRawFile() {} // only called by iflFile::close()

    iflXYZint pageCount;	// number of pages in x,y,z dimensions
    size_t pageSizeBytes;	// number of bytes in each page
    int dataOffset;		// byte offset to first page

    double scaleMinValue;
    double scaleMaxValue;

#if defined(RAW_BUILD_DIO)
    int dioEnab;		// direct I/O enabled flag
#if defined(__sgi)
    int dioAvail;		// direct I/O available flag
    int dioFd;			// direct I/O file descriptor
    struct dioattr dioAttr;	// direct I/O attributes for dioFd
#endif
#endif

    iflStatus decodeArgs(int isOpen);

    int pageIndex(int x, int y, int z, int c) { 
	return ((c/pageSize.c*pageCount.z + z/pageSize.z)*pageCount.y + 
	        y/pageSize.y)*pageCount.x + x/pageSize.x; 
    }
    off64_t pageOffset(int x, int y, int z, int c) {
	return off64_t(pageIndex(x,y,z,c)*off64_t(pageSizeBytes)+dataOffset);
    }

#if !defined(__sgi)
    ssize_t pread(int fd, void* buf, size_t nbytes, off64_t offset)
	{
	    beginFileIO();
	    off64_t rc = lseek64(fd, offset, SEEK_SET);
	    if (rc != offset) return -1;
	    rc = read(fd, buf, nbytes);
	    endFileIO();
	    return (ssize_t)rc;
	}
    ssize_t pwrite(int fd, const void* buf, size_t nbytes, off64_t offset)
	{
	    beginFileIO();
	    off64_t rc = lseek64(fd, offset, SEEK_SET);
	    if (rc != offset) return -1;
	    rc = write(fd, buf, nbytes);
	    endFileIO();
	    return (ssize_t)rc;
	}
#endif
};


class iflRawFormat : public iflFormat {
public:
    iflRawFormat() : iflFormat("Raw") {}
    int typeIsSupported(iflDataType) { return 1; }
    int orderIsSupported(iflOrder) { return 1; }
    int orientationIsSupported(iflOrientation) { return 1; }
    int colorModelIsSupported(iflColorModel) { return 1; }
    int compressionIsSupported(iflCompression cmp) 
	{ return cmp == iflNoCompression; }
    int sizeIsSupported(int,int,int,int, iflOrientation) 
	{ return 1; }
    int pagingIsSupported() { return 1; }
    int pageSizeIsSupported(int,int,int,int, int,int,int,int, iflOrientation) 
	{ return 1; }
    int randomAccessReadIsSupported() { return 1; }
    int randomAccessWriteIsSupported() { return 1; }
    virtual iflDataType getPreferredType() { return iflUChar; }
    virtual iflOrder getPreferredOrder() { return iflInterleaved; }
    iflOrientation getPreferredOrientation() { return iflLowerLeftOrigin; }
    virtual iflColorModel getPreferredColorModel(int nc) 
	{ return nc == 0? iflRGB : iflColorModelFromChans(nc); }
    virtual iflCompression getPreferredCompression() 
	{ return iflNoCompression; }
    void getPreferredPageSize(int sx, int sy, int sz, int sc,
			      int& pWidth, int& pHeight, int& pz, int& pc, 
			      iflOrientation)
	{ pWidth = sx; pHeight = sy; pz = sz; pc = sc; }
    iflFile* newfileobj() { return new iflRawFile; }
};


// This static object declaration causes the file format to be automatically
// registered when the DSO containing this code is dlopen'ed.
//
#if defined(__sgi)
#pragma set woff 1174	// 7.2 compiler doesn't understand /* REFERENCED */
#endif
/* REFERENCED */
static iflRawFormat theRawFormat;
#if defined(__sgi)
#pragma reset woff 1174
#endif


iflStatus
iflRawFile::decodeArgs(int isOpen)
{
    if (isOpen) {
	// fill in default values
	size = iflSize(0, 0, 1, 0);
	pageSize = iflSize(0, 0, 0, 0);
	cm = iflColorModel(0);
	dtype = iflUChar;
	order = iflInterleaved;
	orientation = iflUpperLeftOrigin;
	compression = iflNoCompression;
    }
    dataOffset = 0;

    // pull apart the name value pairs and process them
    const int MaxParams=16;
    iflParameterDesc params[MaxParams];
    int numParams = parseArguments(params, MaxParams);
    
    int minSpecified = FALSE, maxSpecified = FALSE;
    for (int idx = 0; idx < numParams; idx++) {
	// lookup the parameter name
	iflParameterDesc* desc = params+idx;
	int paramID = lookupName(desc->name, paramNames);
	if (paramID == -1) { 
	    iflError(MM_WARNING, "Unknown parameter name: %s", desc->name);
	    continue; 
	}
	
	// decode the value
	if (desc->value == NULL) { 
	    iflError(MM_WARNING, "Missing value for %s parameter", desc->name);
	    continue; 
	}
	
	switch (paramID) {
	
	case SizeParam:
	    sscanf(desc->value, "%d,%d,%d,%d", 
		   &size.x, &size.y, &size.z, &size.c);
	    break;
	
	case PageSizeParam:
	    sscanf(desc->value, "%d,%d,%d,%d", 
		   &pageSize.x, &pageSize.y, &pageSize.z, &pageSize.c);
	    break;

	case ColorModelParam: {
	    int id = lookupName(desc->value, colorModelNames);
	    if (id == -1)
		iflError(MM_WARNING, "Unknown color model: %s", desc->value);
	    else {
		cm = iflColorModel(id);
		if (!isOpen && cm != iflMultiSpectral) 
		    size.c = pageSize.c = iflColorModelChans(cm);
	    }
	    break; }

	case DataTypeParam:
	  { int id = lookupName(desc->value, dataTypeNames);
	    if (id == -1)
		iflError(MM_WARNING, "Unknown data type: %s", desc->value);
	    else
		dtype = iflDataType(id);
	    break;
	  }
	case OrderParam:
	  { int id = lookupName(desc->value, orderNames);
	    if (id == -1)
		iflError(MM_WARNING, "Unknown dimension order: %s", desc->value);
	    else
		order = iflOrder(id);
	    break;
	  }
	case OrientationParam:
	  { int id = lookupName(desc->value, orientationNames);
	    if (id == -1)
		iflError(MM_WARNING, "Unknown orientation: %s", desc->value);
	    else
		orientation = iflOrientation(id);
	    break;
	  }
	case MinimumParam:
	    scaleMinValue = atof(desc->value);
	    minSpecified = TRUE;
	    break;
	    
	case MaximumParam:
	    scaleMaxValue = atof(desc->value);
	    maxSpecified = TRUE;
	    break;
	    
	case OffsetParam:
	    dataOffset = iflMax(0, atoi(desc->value));
	    break;
	}
    }

    if (isOpen) {
	// process image size, color model defaults
	if (size.x <= 0 || size.y <= 0 || size.c < 0 || size.z < 0)
	    return iflStatusEncode(iflBADPARAMS);
	if (!size.z) size.z = 1;
	if (cm == iflColorModel(0)) 
	    cm = size.c < 1? iflRGB : iflColorModelFromChans(size.c);
	if (!size.c) {
	    if (cm == iflMultiSpectral) return iflStatusEncode(iflBADPARAMS);
	    size.c = iflColorModelChans(cm);
	}
	
	// process pags size default
	if (pageSize.x < 0 || pageSize.y < 0 || 
	    pageSize.c < 0 || pageSize.z < 0)
	    return iflStatusEncode(iflBADPARAMS);
	if (pageSize.x == 0) pageSize.x = size.x;
	if (pageSize.y == 0) pageSize.y = size.y;
	if (pageSize.z == 0) pageSize.z = 1;
	if (pageSize.c == 0) 
	    pageSize.c = order == iflSeparate? 1 : size.c;
    }
    
    // deal with scale min max defaults
    if (!minSpecified) 
	scaleMinValue = iflDataIsIntegral(dtype)? iflDataMin(dtype) : 0;
    if (!maxSpecified)
	scaleMaxValue = iflDataIsIntegral(dtype)? iflDataMax(dtype) : 1;

    // deal with transposed orientations
    if (iflOrientationIsTrans[orientation]) {
	int tmp;
	tmp = size.x;
	size.x = size.y;
	size.y = tmp;
	tmp = pageSize.x;
	pageSize.x = pageSize.y;
	pageSize.y = tmp;
    }

    // calculate page related paramters
    pageCount.x = (size.x-1)/pageSize.x + 1;
    pageCount.y = (size.y-1)/pageSize.y + 1;
    pageCount.z = (size.z-1)/pageSize.z + 1;
    pageSizeBytes = iflDataSize(dtype, pageSize);

    return iflOKAY;
}


iflStatus iflRawFile::createFile()
{
    // decode the format args
    iflStatus status = decodeArgs(FALSE);
    if (status != iflOKAY) return status;

    if (fd < 0) {
	assert(filename != NULL);
#	if defined(_WIN32)
	fd = ::open(filename, accessmode|O_CREAT|O_TRUNC|O_BINARY, 0666);
#	else
	fd = ::open(filename, accessmode|O_CREAT|O_TRUNC, 0666);
#	endif
	if (fd < 0)
	    return iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX, oserror());
    } else {
#	ifndef _WIN32
	(void)::ftruncate(fd, (off_t)0);
#	endif
    }

#if defined(RAW_BUILD_DIO)
#if defined(__sgi)
    if (pageSizeBytes >= RAW_DIO_MIN_PAGESIZE) {
	dioFd = (fcntl(fd, F_DIOINFO, &dioAttr) != -1) ?
		::open(filename, accessmode|O_DIRECT) : -1;

	dioAvail = dioEnab = dioFd != -1;
    } else {
	dioEnab = dioAvail = FALSE;
	dioFd = -1;
    }
#else
    dioEnab = FALSE;
#endif
#endif

    return iflOKAY;
}


iflStatus iflRawFile::openFile()
{
    // decode the format args
    iflStatus status = decodeArgs(TRUE);
    if (status != iflOKAY) return status;

    if (fd < 0) {
	assert(filename != NULL);
#	if defined(_WIN32)
	fd = ::open(filename, accessmode|O_BINARY);
#	else
	fd = ::open(filename, accessmode);
#	endif
	if (fd < 0)
	    return iflStatusEncode(iflOPENFAILED, iflSubDomainUNIX, oserror());
    }

#if defined(RAW_BUILD_DIO)
#if defined(__sgi)
    if (pageSizeBytes >= RAW_DIO_MIN_PAGESIZE) {
	dioFd = (fcntl(fd, F_DIOINFO, &dioAttr) != -1) ?
		::open(filename, accessmode|O_DIRECT) : -1;

	dioAvail = dioEnab = dioFd != -1;
    } else {
	dioEnab = dioAvail = FALSE;
	dioFd = -1;
    }
#else
    dioEnab = FALSE;
#endif
#endif

    return iflOKAY;
}


// If this gets called, it's safe to
// assume that openFile() has been called and it succeeded.
//
iflStatus iflRawFile::closeFile()
{
    assert(fd >= 0);

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    if (dioFd != -1) {
	::close(dioFd);
	dioFd = -1;
    }
#endif

    if (::close(fd) != 0)
	return iflStatusEncode(iflCLOSEFAILED, iflSubDomainUNIX, oserror());

    return iflOKAY;
}


iflStatus 
iflRawFile::getPage(void* data, int x, int y, int z, int c, int,int,int,int)
{
    off64_t pgsoff = pageOffset(x,y,z,c);

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    // Can't do direct IO if not enabled, buffer not cache aligned or
    // page not cache aligned
    if (!dioEnab || !dioAvail || (long)data % dioAttr.d_mem != 0 ||
	    pgsoff % dioAttr.d_mem != 0 || pageSizeBytes < dioAttr.d_miniosz)
#endif
	return PREAD(fd, data, pageSizeBytes, pgsoff) != -1 ? iflOKAY :
		iflStatusEncode(iflREADFAILED, iflSubDomainUNIX, oserror());

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    // Read any portion that is not block aligned
    off64_t pgeoff = pgsoff + pageSizeBytes;
    off_t slen = dioAttr.d_miniosz - (pgsoff % dioAttr.d_miniosz);
    off_t elen = pgeoff % dioAttr.d_miniosz;
    uchar_t* cdata = (uchar_t*)data;
    ssize_t sts;
    if (elen > 0) {
	sts = PREAD(fd, cdata + (pgeoff - elen - pgsoff), elen,
		pgeoff - elen);
	if (sts == -1)
	    return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
		    oserror());

	pgeoff -= elen;
    }

    if (slen > 0) {
	sts = PREAD(fd, data, slen, pgsoff);
	if (sts == -1)
	    return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
		    oserror());
	pgsoff += slen;
	cdata += slen;
    }

    off64_t len = pgeoff - pgsoff;
    if (len > 0) {
	while (len > 0) {
	    sts = PREAD(dioFd, cdata, len % dioAttr.d_maxiosz, pgsoff);
	    if (sts == -1)
		return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX,
			oserror());
	    len -= dioAttr.d_maxiosz;
	    pgsoff += dioAttr.d_maxiosz;
	    cdata += dioAttr.d_maxiosz;
	}
    }

    return iflOKAY;
#endif
}


iflStatus 
iflRawFile::setPage(const void* data, int x,int y,int z,int c,int,int,int,int)
{
    off64_t pgsoff = pageOffset(x,y,z,c);

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    // Can't do direct IO if not enabled, buffer not cache aligned or
    // page not cache aligned
    if (!dioEnab || !dioAvail || (long)data % dioAttr.d_mem != 0 ||
	    pgsoff % dioAttr.d_mem != 0 || pageSizeBytes < dioAttr.d_miniosz)
#endif
	return PWRITE(fd, data, pageSizeBytes, pgsoff) != -1 ? iflOKAY :
		iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX, oserror());

#if defined(__sgi) && defined(RAW_BUILD_DIO)
    // Write any portion that is not block aligned
    off64_t pgeoff = pgsoff + pageSizeBytes;
    off64_t slen = dioAttr.d_miniosz - (pgsoff % dioAttr.d_miniosz);
    off64_t elen = pgeoff % dioAttr.d_miniosz;
    uchar_t* cdata = (uchar_t*)data;
    ssize_t sts;
    if (elen > 0) {
	sts = PWRITE(fd, cdata + (pgeoff - elen - pgsoff), elen,
		pgeoff - elen);
	if (sts == -1)
	    return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
		    oserror());
	pgeoff -= elen;
    }

    if (slen > 0) {
	sts = PWRITE(fd, data, slen, pgsoff);
	if (sts == -1)
	    return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
		    oserror());
	pgsoff += slen;
	cdata += slen;
    }

    off64_t len = pgeoff - pgsoff;
    if (len > 0) {
	while (len > 0) {
	    sts = PWRITE(dioFd, cdata, len % dioAttr.d_maxiosz, pgsoff);
	    if (sts == -1)
		return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX,
			oserror());
	    len -= dioAttr.d_maxiosz;
	    pgsoff += dioAttr.d_maxiosz;
	    cdata += dioAttr.d_maxiosz;
	}
    }

    return iflOKAY;
#endif
}


iflStatus 
iflRawFile::getScaleMinMax(double& min, double& max)
{
    min = scaleMinValue;
    max = scaleMaxValue;
    return iflOKAY;
}


iflStatus 
iflRawFile::setScaleMinMax(double min, double max)
{
    scaleMinValue = min;
    scaleMaxValue = max;
    return iflOKAY;
}


#if defined(RAW_BUILD_DIO)
iflStatus
iflRawFile::_enableDirectIO(int enable)
{
    if (dioAvail)
	dioEnab = enable != 0;
    return iflOKAY;
}


iflStatus
iflRawFile::_getDirectIOMemAlignment(size_t& memAlign)
{
#if defined(__sgi)
    memAlign = dioAvail ? dioAttr.d_mem : 0;
#else
    memAlign = 1;
#endif
    return iflOKAY;
}
#endif
