/*
    TIFF i/o driver
*/
#include <ifl/iflFormat.h>
#include <ifl/iflFile.h>
#include <ifl/iflDataSize.h>
#include <ifl/iflSGIColormap.h>
#include <ifl/iflColor.h>
#include <fcntl.h>
#if defined(_WIN32)
#include <windows.h>
#include <io.h>
#include <malloc.h>
#define oserror() errno
#else
#include <unistd.h>
#include <alloca.h>
#endif
#include <string.h>
#include <errno.h>
#include "tiffio.h"
#include "tiffiop.h"
#include "xtiffio.h"
#include "xtiffiop.h"


class iflTIFFFile : public iflFile {
public:
    virtual int getNumImgs()	{ return nimgs; }
    virtual int getCurrentImg()	{ return imgidx; }
    virtual iflStatus setCurrentImg(int);
    virtual iflStatus appendImg(iflFile* source, iflFileConfig* fc=NULL);
    virtual iflStatus flush();
    virtual iflStatus getPage(void* data, int,int,int,int, int,int,int,int);
    virtual iflStatus setPage(const void* data, int,int,int,int,
			      int,int,int,int);

    virtual iflStatus getScaleMinMax(double& min, double& max);
    virtual iflStatus getStatMinMax(double& min,double& max);
    virtual iflStatus setStatMinMax(double min, double max);
    virtual iflStatus getColormap(const iflColormap*&);
    virtual iflStatus setColormap(const iflColormap*);
    virtual iflStatus getItem(int tag, va_list);
    virtual iflStatus setItem(int tag, va_list);

private:
    virtual iflStatus openFile();
    virtual iflStatus createFile();
    virtual iflStatus closeFile();
    iflStatus _getICCProfile(int &size, void*& profile);
    virtual iflStatus _freeICCProfile(void* profile);
    iflStatus _setICCProfile(int size, const void* profile);

    friend class iflTIFFFormat;
    iflTIFFFile() {}  // only called by iflTIFFFormat::newfileobj()
    ~iflTIFFFile() {} // only called by iflFile::close()

    struct tiff* tiff;			// point to the current directory?
    char* tag;

    size_t pageSizeBytes;
    int nimgs;				// number of images in the file
    int imgidx;				// current image index in file
    iflColormap* cmap;
    int writeInit;  	    	    	// do we need to init the write buffers?

    void initErrorHandler();
    iflStatus initDir();
    iflStatus initColorModel();
    iflStatus checkColorModel();


    iflStatus initImg(const iflSize& sz, iflDataType tp, iflOrder o,
		      iflOrientation orientation, const iflSize& psz,
		      iflCompression cp, iflColorModel col);
};


class iflTIFFFormat : public iflFormat {
public:
    iflTIFFFormat() : iflFormat("TIFF") {}
    int typeIsSupported(iflDataType) { return 1; }
    int orderIsSupported(iflOrder order)
	{
	    return order == iflInterleaved
		|| order == iflSeparate;
	}
    int orientationIsSupported(iflOrientation) { return 1; }
    int colorModelIsSupported(iflColorModel cm)
	{
	    return cm == iflNegative
		|| cm == iflLuminance
		|| cm == iflLuminanceAlpha
		|| cm == iflRGB
		|| cm == iflRGBPalette
		|| cm == iflCMY
		|| cm == iflCMYK
		|| cm == iflRGBA;
	}
    int compressionIsSupported(iflCompression compress)
	{
	    return compress == iflNoCompression
		|| compress == iflCCITTFAX3
		|| compress == iflCCITTFAX4
		|| compress == iflLZW
		|| compress == iflPACKBITS
		|| compress == iflJPEG;
	}
    int sizeIsSupported(int,int,int,int, iflOrientation) 
	{ return 1;}
    int pagingIsSupported() { return 1; }
    int pageSizeIsSupported(int iWidth, int /*iHeight*/, int /*iz*/, int /*ic*/,
			    int width, int height, int /*z*/, int /*c*/,
			    iflOrientation)
        {
	    return width == iWidth || ((width&0xf) == 0 && (height&0xf) == 0);
	}
    int randomAccessReadIsSupported() { return 1; }
    int randomAccessWriteIsSupported() { return 0; }
    virtual iflDataType getPreferredType() { return iflUChar; }
    virtual iflOrder getPreferredOrder() { return iflInterleaved; }
    iflOrientation getPreferredOrientation() { return iflUpperLeftOrigin; }
    virtual iflColorModel getPreferredColorModel(int nc)
	{ return nc == 0? iflRGB : iflColorModelFromChans(nc); }
    virtual iflCompression getPreferredCompression() 
	{ return iflNoCompression; }
    void getPreferredPageSize(int, int, int, int ic,
			      int& pWidth, int& pHeight, int& pz, int& pc, 
			      iflOrientation)
	{ pWidth = 128; pHeight = 128; pz = 1; pc = ic; }
    iflFile* newfileobj() { return new iflTIFFFile; }
    virtual iflTagDescription* _getTagDescriptions(int& count);
};


iflTagDescription*  
iflTIFFFormat::_getTagDescriptions(int& count)
{
    static iflEnumSpec resSpec[] = {
    	{ "None", RESUNIT_NONE },
    	{ "Inch", RESUNIT_INCH },
    	{ "Centimeter", RESUNIT_CENTIMETER },
    	{ NULL, 0 }
    };

    static iflTagDescription desc[] = {
      { TIFFTAG_IMAGEDESCRIPTION, 
    	iflTagGet|iflTagSet|iflTagString|iflTagOneParam|iflTagCreate|iflTagInfo,
    	"A description of the subject of the image",
    	"Image Description",
    	"",
    	NULL },
      { TIFFTAG_ARTIST, 
    	iflTagGet|iflTagSet|iflTagString|iflTagOneParam|iflTagCreate|iflTagInfo,
    	"The name of the artist who created the image",
    	"Artist",
    	"",
    	NULL },
      { TIFFTAG_RESOLUTIONUNIT,
    	iflTagGet|iflTagSet|iflTagShortEnum|iflTagOneParam
    	|iflTagCreate|iflTagInfo,
    	"The unit of measurement for X and Y resolution",
    	"Resolution unit",
    	"None",
    	&resSpec },
      { TIFFTAG_XRESOLUTION, 
    	iflTagGet|iflTagSet|iflTagFloat|iflTagOneParam|iflTagCreate|iflTagInfo,
    	"Number of pixels per resolution unit in image width direction",
    	"X resolution",
    	NULL,
    	NULL },
      { TIFFTAG_YRESOLUTION, 
    	iflTagGet|iflTagSet|iflTagFloat|iflTagOneParam|iflTagCreate|iflTagInfo,
    	"Number of pixels per resolution unit in image height direction",
    	"Y resolution",
    	NULL,
    	NULL }
    };

    count = sizeof desc / sizeof (iflTagDescription);
    return desc;
}


// This static object declaration causes the file format to be automatically
// registered when the DSO containing this code is dlopen'ed.
//
#if defined(__sgi)
#pragma set woff 1174	// 7.2 compiler doesn't understand /* REFERENCED */
#endif
/* REFERENCED */
static iflTIFFFormat theTIFFformat;
#if defined(__sgi)
#pragma reset woff 1174
#endif

static int ifl_tiff_align(TIFF*, u_char*, int);
static void ifl_tiff_bitshift(const void* ibuf, void* obuf,
			      int nbits, int nlines);
static void ifl_tiff_bitcopy(const void* ibuf, void* obuf,
			     int startbit, int nbits);
static void ifl_tiff_bitstobytes(const void* ibuf, void* obuf,
				 int bs, int cnt);

const int DefaultPgSz = 0x10000;    // Default page size in bytes (max)

#define TIFF_BadPixType(type) (type>iflDouble)
#define BytesPerChannel ((tiff->tif_dir.td_bitspersample+7)/8)
#define BadChanBnds(c) (c<0 || c>tiff->tif_dir.td_samplesperpixel) ? TRUE:FALSE

// XXX module name is lost, oh well
static void
myTIFFErrorHandler(const char * /*module*/, const char *fmt, va_list ap)
{
    // even though it's an error, it's probably not worth aborting over
    iflErrorV(MM_WARNING, fmt, ap);
}
static void
myTIFFWarningHandler(const char * /*module*/, const char *fmt, va_list ap)
{
    if (strncmp("unknown field with tag ", fmt, 23) != 0)
    	iflErrorV(MM_WARNING, fmt, ap);
}

void
iflTIFFFile::initErrorHandler()
{
    static int setYet = FALSE;
    
    if (!setYet) {
	// silence error and warning reporting
	TIFFSetErrorHandler(myTIFFErrorHandler);
	TIFFSetWarningHandler(myTIFFWarningHandler);
	setYet = TRUE;
    }
}

inline const char *open_mode_to_fopen_mode(int mode)
{
    return
	mode == O_RDONLY ? "rm" :
	mode == O_WRONLY ? "w" :
	mode == O_RDWR ? "r+" :
	NULL;
}

/*
 * Open a TIFF file and initialize iflFile data attributes
 */
iflStatus
iflTIFFFile::openFile()
{
    tiff = NULL;	// so we can abort and call destructor at any time
    cmap = NULL;
    writeInit = FALSE;

    initErrorHandler();

    if (fd >= 0) {
	if (lseek(fd, (off_t)0, SEEK_SET) == (off_t)-1)
	    return iflStatusEncode(iflSEEKFAILED, iflSubDomainUNIX,oserror());
	tiff = XTIFFFdOpen(fd,filename,open_mode_to_fopen_mode(accessmode));
    } else {
	iflAssert(filename != NULL); // caller guarantees this
	tiff = XTIFFOpen(filename, open_mode_to_fopen_mode(accessmode));
    }
    if (tiff != NULL)
    {
	/*
	 * Populate the iflFile data slots
	 */
	/* get the number of images in the file */
	nimgs = 1; imgidx = 0;
	while (TIFFReadDirectory(tiff) != 0) nimgs++;
	TIFFSetDirectory(tiff, imgidx); /* reset to first image */

	/*
 	 * Initialize the ilImage slots based on the tiff structure
	 */
	return initDir();
    }
    else
	return iflStatusEncode(iflBADIMGFMT);
}

/*
 * Get ifl fields from tiff struct
 */
iflStatus
iflTIFFFile::initDir()
{
    u_short val;
    int bpp;


	/*
 	 * Populate the ilImage data slots
	 */
	size.x = (int)tiff->tif_dir.td_imagewidth; 
	size.y = (int)tiff->tif_dir.td_imagelength; 
	size.z = (int)tiff->tif_dir.td_imagedepth;
	size.c = tiff->tif_dir.td_samplesperpixel; 
#ifdef DEFUNCT
	tag = tiff->tif_dir.td_imagedescription;
#endif
	order = (tiff->tif_dir.td_planarconfig == PLANARCONFIG_SEPARATE) ?
		iflSeparate : iflInterleaved;
	orientation = (iflOrientation) tiff->tif_dir.td_orientation;
        if (orientation == 0) orientation = iflUpperLeftOrigin;


	if (tiff->tif_dir.td_sampleformat == SAMPLEFORMAT_IEEEFP){
	    dtype = (tiff->tif_dir.td_bitspersample == sizeof(float)*8) ? 
			iflFloat : iflDouble;
	}
	else if (tiff->tif_dir.td_bitspersample == 1)  {
	    dtype = iflBit;
	}
	else {
	    bpp = (tiff->tif_dir.td_bitspersample+7) / 8;
	    if (bpp <= sizeof(char))
		 dtype = tiff->tif_dir.td_sampleformat == SAMPLEFORMAT_INT ?
		      	    iflChar : iflUChar;
	    else if (bpp <= sizeof(short))
		 dtype = tiff->tif_dir.td_sampleformat == SAMPLEFORMAT_INT ?
		      	    iflShort : iflUShort;
	    else 
		 dtype = tiff->tif_dir.td_sampleformat == SAMPLEFORMAT_INT ?
		      	    iflInt : iflUInt;
	}

	/* If it's a standard TIFF file, then go by rowsperstrip */
	if (isTiled(tiff)) {
	    pageSize.y = (int)tiff->tif_dir.td_tilelength;
	    pageSize.x = (int)tiff->tif_dir.td_tilewidth;
	}
	else {
	    pageSize.y = (tiff->tif_dir.td_rowsperstrip == 0xffffffffL) ?
			 size.y : (int)tiff->tif_dir.td_rowsperstrip;
	    pageSize.x = (int)tiff->tif_dir.td_imagewidth;
	}


	if (!TIFFFieldSet(tiff, FIELD_TILEDEPTH))
	    pageSize.z = (int)tiff->tif_dir.td_imagedepth;
	else
	    pageSize.z = (int)tiff->tif_dir.td_tiledepth;
	pageSize.c = order == iflSeparate? 1:size.c;
	if (dtype == iflBit && !isTiled(tiff)) 
	    pageSizeBytes = pageSize.c*pageSize.y*pageSize.z*
			    TIFFScanlineSize(tiff);
	else
	    pageSizeBytes = iflDataSize(dtype, pageSize);

    // Set compression
    //
    if (!TIFFGetField(tiff, TIFFTAG_COMPRESSION, &val)) 
    	return iflStatusEncode(iflBADFIELDGET);
    else switch (val) {
	case COMPRESSION_NONE:      compression = iflNoCompression; break;
	case COMPRESSION_CCITTFAX3: compression = iflCCITTFAX3;     break;
	case COMPRESSION_CCITTFAX4: compression = iflCCITTFAX4;     break;
	case COMPRESSION_LZW:       compression = iflLZW;           break;
	case COMPRESSION_PACKBITS:  compression = iflPACKBITS;      break;
	case COMPRESSION_JPEG:      compression = iflJPEG;          break;
	default:    	    	    compression = iflUnknownCompression; break;
    }
    
    // Set color model
    switch (tiff->tif_dir.td_photometric) {
	case PHOTOMETRIC_MINISWHITE:
		cm = iflNegative;
		break;
	case PHOTOMETRIC_MINISBLACK:
		cm = iflLuminance;
		break;
	case PHOTOMETRIC_RGB:
		cm = iflRGB;
		break;
	case PHOTOMETRIC_PALETTE:
		cm = iflRGBPalette;
		break;
	case PHOTOMETRIC_SEPARATED:
		cm = (size.c == 4) ? iflCMYK : iflCMY;
		break;
	case PHOTOMETRIC_YCBCR:
		if (compression == iflJPEG) {
		    if (!TIFFSetField(tiff, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB))
			return iflStatusEncode(iflBADFIELDSET);
		    cm = iflRGB;
		    break;
		}
		/* FALLTHROUGH */
	case PHOTOMETRIC_CIELAB:
	default:
		iflError(MM_ERROR, "%s has unsupported TIFF color model %d\n",
			 filename, tiff->tif_dir.td_photometric);
		return iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
    }

    // check for "extra" channels
    //
    int extra = tiff->tif_dir.td_extrasamples;
    if (extra > 1)
	return iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
    else if (extra == 1) {
	
	// are they some sort of alpha
	u_short extraType = *tiff->tif_dir.td_sampleinfo;
	int assoc = extraType == EXTRASAMPLE_ASSOCALPHA;
	if (assoc || extraType == EXTRASAMPLE_UNASSALPHA) {

	    // we only do alpha for luminance and rgb for now
	    switch (cm) {
	    case iflLuminance:	
		cm = iflLuminanceAlpha; 
		break;
	    case iflRGB:
		cm = iflRGBA; 
		break;
	    default:
		return iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	    }
	    
	    if (assoc)
		iflError(MM_WARNING, 
			 "%s: associated alpha treated as unassociated", 
			 filename);
	}
    }
    
    // check for bogus 4 channel RGB (without extrasamples set) and allow
    // it to be treated as RGBA
    
    if (size.c == 4 && cm == iflRGB) {
	cm = iflRGBA;
	iflError(MM_WARNING, 
		 "%s: assuming four channel RGB image is really RGBA", 
		 filename);
    }
    

    // make sure the color model is ok
    iflStatus sts = checkColorModel();
    if (sts != iflOKAY)	return sts;

    // intialize color map
    //
    delete cmap; 
    cmap = NULL; 
    if (cm == iflRGBPalette ) {
	u_short *r, *g, *b;
	if (!TIFFGetField(tiff, TIFFTAG_COLORMAP, &r, &g, &b)) {
	    if (dtype > iflChar)  {
		// XXX should emit error message here; the file is bogus
		return iflStatusEncode(iflBADIMGFMT);
	    }
	    cmap = new iflSGIColormap;
        } else {
	    int len = int(1L<<tiff->tif_dir.td_bitspersample);
	    
	    // hack to treat weird images with color map that only has
	    // 0-255 range as having uchar cmap instead of ushort
	    int max = 0;
	    for (int idx = 0; idx < len; idx++) {
	    	int rv = r[idx], gv = g[idx], bv = b[idx];
	    	max = iflMax(max, rv, gv, bv);
	    }
	    if (max < 256) {
	    	iflError(MM_WARNING, "%s: assuming 8-bit colormap", filename);
		cmap = new iflColormap(3, iflUChar, 0, len-1);
		u_char* rm = (u_char*)cmap->getChan(0);
		u_char* gm = (u_char*)cmap->getChan(1);
		u_char* bm = (u_char*)cmap->getChan(2);
	    	for (idx = 0; idx < len; idx++) { 
		    rm[idx] = (u_char)r[idx]; 
		    gm[idx] = (u_char)g[idx]; 
		    bm[idx] = (u_char)b[idx];
		}
	    }
	    else {
		cmap = new iflColormap(3, iflUShort, 0, len-1);
		memcpy(cmap->getChan(0), r, len*sizeof(u_short));
		memcpy(cmap->getChan(1), g, len*sizeof(u_short));
		memcpy(cmap->getChan(2), b, len*sizeof(u_short));
	    }
	    
        }
    }

#ifdef DEFUNCT /* want to support something like this? */
    // set description if any
    char* desc;
    if (TIFFGetField(tiff, TIFFTAG_IMAGEDESCRIPTION, &desc))
	setDescription(desc);
#endif
    
    return iflOKAY;
}


iflStatus
iflTIFFFile::getScaleMinMax(double& min, double& max)
{
    // using bits-per-sampe for floating point data just doesn't
    // make any sense
    
    if (!iflDataIsIntegral(dtype)) return getStatMinMax(min, max);

    // but for integral types, that's what we should use, not the
    // scale min/max values which aren't supposed to be use to 
    // affect display

    int bps = tiff->tif_dir.td_bitspersample;

    if (iflDataIsSigned(dtype)) {
	min = -(1 << (bps-1));
	max = (1 << (bps-1)) - 1;
    }
    else {
	min = 0;
	max = (1 << bps) - 1;
    }
    
    return iflOKAY;
}


iflStatus
iflTIFFFile::getStatMinMax(double& statMin, double& statMax)
{
    // Get the min and max pixel values
    //
    // Note, MIN and MAX values can be of different
    // types than the image data type!
    // There are actually 2 TIFF tags that can be used,
    // one is hard coded to be SHORTS the other is more flexible.
    // Default to the more flexible better, otherwise use the other.
    //
    short sval;
    unsigned short uval;
    if (!TIFFGetField(tiff, TIFFTAG_SMINSAMPLEVALUE, &statMin)) {
	if (dtype == iflChar || dtype == iflShort || dtype == iflInt) {
	    if (!TIFFGetField(tiff, TIFFTAG_MINSAMPLEVALUE, &sval))
		sval = -(1<<(tiff->tif_dir.td_bitspersample-1));
	    statMin = sval;
	}
	else {
	    if (!TIFFGetField(tiff, TIFFTAG_MINSAMPLEVALUE, &uval))
		uval = 0;
	    statMin = uval;
	}
    }

    if (!TIFFGetField(tiff, TIFFTAG_SMAXSAMPLEVALUE, &statMax)) {
	if (dtype == iflChar || dtype == iflShort || dtype == iflInt) {
	    if (!TIFFGetField(tiff, TIFFTAG_MAXSAMPLEVALUE, &sval))
		sval = (1<<(tiff->tif_dir.td_bitspersample-1))-1;
	    statMax = sval;
	}
	else {
	    if (!TIFFGetField(tiff, TIFFTAG_MAXSAMPLEVALUE, &uval))
		uval = (1<<tiff->tif_dir.td_bitspersample)-1;
	    statMax = uval;
	}
    }
    return iflOKAY;
}

iflStatus
iflTIFFFile::createFile()
{
    tiff = NULL;	// so we can abort and call destructor at any time
    cmap = NULL;
    writeInit = FALSE;

    initErrorHandler();

    if ((tiff = XTIFFOpen(filename, "w")) == (TIFF *)0 )
	return iflStatusEncode(iflOPENFAILED);

    nimgs = 1;
    imgidx = 0;

    iflStatus status = initImg(size, dtype, order, orientation, 
			       pageSize, compression, cm);
    if (status != iflOKAY) {
	XTIFFClose(tiff); // XXX wrong-- should keep fd open unless we opened it
	tiff = (TIFF *)0;
	return status;
    }
    return iflOKAY;
}

// set tiff struct stuff from ifl fields
iflStatus
iflTIFFFile::initImg(const iflSize& sz, iflDataType type, iflOrder o, 
		     iflOrientation orient, const iflSize& psz, 
		     iflCompression cmprs, iflColorModel col)
{
    iflStatus status = iflOKAY;

    // Set the image attributes, data type and write to the header.
    //
    size = sz;
    dtype = type;
    order = o;
    orientation = orient;
    compression = cmprs;
    TIFFFreeDirectory(tiff);
    TIFFDefaultDirectory(tiff);
    tiff->tif_flags &= ~TIFF_ISTILED;
    tiff->tif_diroff = 0;

    if (!TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, sz.x) ||
	!TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, sz.y) ||
	sz.z != 1 && !TIFFSetField(tiff, TIFFTAG_IMAGEDEPTH, sz.z) ||
	TIFF_BadPixType(type)) {
	status = iflStatusEncode(iflBADFIELDSET);
	return status;
    }

    // Since all's okay, set the samples per pixel and bits per channel
    //
    int bpp = (int)iflDataSize(type, 8);
    int ord = (o==iflInterleaved) ? PLANARCONFIG_CONTIG : PLANARCONFIG_SEPARATE;
    int ttype;
    switch (type) {
    case iflBit:
    case iflUChar:
    case iflUShort:
    case iflUInt:
	ttype = SAMPLEFORMAT_UINT; break;
    case iflChar:
    case iflShort:
    case iflInt:
	ttype = SAMPLEFORMAT_INT; break;
    case iflFloat:
    case iflDouble:
	ttype = SAMPLEFORMAT_IEEEFP; break;
    default:
	ttype = SAMPLEFORMAT_VOID; break;
    }

    if (!TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, bpp) ||
	!TIFFSetField(tiff, TIFFTAG_SAMPLEFORMAT, ttype) ||
	!TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, ord) ||
	// IFL and TIFF happen to use the exact same orientation coding
	!TIFFSetField(tiff, TIFFTAG_ORIENTATION,orient) ||
	!TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, sz.c)) 
    {
	status = iflStatusEncode(iflBADFIELDSET);
	return status;
    }

    // Set compression
    int compress;
    switch(compression) {
	case iflNoCompression: compress = COMPRESSION_NONE;      break;
	case iflCCITTFAX3:     compress = COMPRESSION_CCITTFAX3; break;
	case iflCCITTFAX4:     compress = COMPRESSION_CCITTFAX4; break;
	case iflLZW:           compress = COMPRESSION_LZW;       break;
	case iflPACKBITS:      compress = COMPRESSION_PACKBITS;  break;
	case iflJPEG:          compress = COMPRESSION_JPEG;      break;
	default: assert(0); break;
    }
    if (!TIFFSetField(tiff, TIFFTAG_COMPRESSION, compress))
	return iflStatusEncode(iflBADFIELDSET);

    // Attempt to specify photometric interpretation
    if (col == (iflColorModel)0) {
        if ((status = initColorModel()) != iflOKAY)
	    return status;;
    }
    else {
    	// Make sure that the number of channels coincides
    	// with the given color model 
	cm = col;
	if ((status = checkColorModel()) != iflOKAY)
	    return status;
    }

    switch(cm) {
    case iflNegative: 
	if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE))
	    return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflLuminanceAlpha:
    case iflLuminance: 
	if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK))
	    return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflRGBA: 
    case iflRGB: 
	if (compression == iflJPEG) {
	    if (   !TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_YCBCR)
		|| !TIFFSetField(tiff, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB))
		return iflStatusEncode(iflBADFIELDSET);
	} else
	    if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB))
		return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflRGBPalette: 
	if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE))
	    return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflCMY: 
	if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC ,PHOTOMETRIC_SEPARATED))
	    return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflCMYK: 
	if (!TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_SEPARATED))
	    return iflStatusEncode(iflBADFIELDSET);
	break;
    default:
	iflError(MM_ERROR, "Bad color model %d", cm);
	return iflStatusEncode(iflBADPARAMS);
    }

    // handle alpha
    if (cm == iflLuminanceAlpha || cm == iflRGBA) {
	uint16 extrasamples[1] = { EXTRASAMPLE_UNASSALPHA };
	if (!TIFFSetField(tiff, TIFFTAG_EXTRASAMPLES, 1, extrasamples))
	return iflStatusEncode(iflBADFIELDSET);
    }

    // Set orientation
    if (!TIFFSetField(tiff, TIFFTAG_ORIENTATION, orientation))
	return iflStatusEncode(iflBADFIELDSET);

    // Set the strip or tile parameters
    pageSize.z = psz.z <= 0? 1:psz.z;
    pageSize.c = order == iflSeparate? 1:size.c;
    if ((pageSize.x = psz.x) <= 0) pageSize.x = size.x;
    if ((pageSize.y = psz.y) <= 0) {
	pageSize.y = DefaultPgSz/(int)iflDataSize(dtype, pageSize.x*pageSize.z*
						         pageSize.c);
	if (pageSize.y > size.y) pageSize.y = size.y;
    }

    tiff->tif_dir.td_stripsperimage = (pageSize.y+size.y-1) / pageSize.y;
    if (pageSize.x == size.x && pageSize.z == 1 && size.z == 1) {
	if (!TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, pageSize.y)) {
	    status = iflStatusEncode(iflBADFIELDSET);
	    return status;
	}
    }
    else {
	// make sure the x and y page sizes conform to TIFF requirements
	pageSize.x = (pageSize.x+15) & ~15;
	pageSize.y = (pageSize.y+15) & ~15;
	if (!TIFFSetField(tiff, TIFFTAG_TILEWIDTH, pageSize.x)) {
	    status = iflStatusEncode(iflBADFIELDSET);
	    return status;
	}
	if (!TIFFSetField(tiff, TIFFTAG_TILELENGTH, pageSize.y)) {
	    status = iflStatusEncode(iflBADFIELDSET);
	    return status;
	}
	if (!TIFFSetField(tiff, TIFFTAG_TILEDEPTH, pageSize.z)) {
	    status = iflStatusEncode(iflBADFIELDSET);
	    return status;
	}
    }
    tiff->tif_scanlinesize = TIFFScanlineSize(tiff);
    if (dtype == iflBit && !isTiled(tiff)) 
      pageSizeBytes = pageSize.c*pageSize.y*pageSize.z* TIFFScanlineSize(tiff);
    else
      pageSizeBytes = iflDataSize(dtype, pageSize);
    tiff->tif_tilesize = (tsize_t)pageSizeBytes;

    // advertise our selves
    TIFFSetField(tiff, TIFFTAG_SOFTWARE, "SGI's Image Format Library/1.2");

    return status;
}


iflStatus
iflTIFFFile::initColorModel()
{
    switch (size.c) {
	case 1:  cm = iflLuminance; break;
	case 2:  cm = iflLuminanceAlpha; break;
	case 3:  cm = iflRGB; break;
	case 4:  cm = iflRGBA; break; 
	default: return iflStatusEncode(iflBADFIELDSET);
    }
    return iflOKAY;
}


iflStatus
iflTIFFFile::checkColorModel()
{
    switch (cm) {
    case iflNegative:
    case iflLuminance:
    case iflRGBPalette:
	if (size.c != 1) return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflLuminanceAlpha:
	if (size.c != 2) return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflRGB:
	if (size.c != 3) return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflRGBA:
	if (size.c != 4) return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflCMY:
	if (size.c != 3) return iflStatusEncode(iflBADFIELDSET);
	break;
    case iflCMYK:
	if (size.c != 4) return iflStatusEncode(iflBADFIELDSET);
	break;
    default:
	return initColorModel();
    }
    return iflOKAY;
}


iflStatus
iflTIFFFile::closeFile()	/* clean up when done */
{
    if (cmap != NULL)
	delete cmap;
    if (tiff)
	XTIFFClose(tiff);
    return iflOKAY;
}

iflStatus 
iflTIFFFile::getPage(void* data, int x, int y, int z, int c, int,int,int,int)
{
    register TIFFDirectory *td;
    int tile;
    size_t bytecount;
    
    // If the file's open for reading then just return
    if (accessmode == O_WRONLY)
	return iflStatusEncode(iflWRITEONLY);

    // If there's nothing to get from just return.
    if (tiff == NULL || &tiff->tif_dir == NULL || 
	tiff->tif_dir.td_stripbytecount== NULL) {
	  return iflOKAY;
    }

    // Try and get the tile required to fill the tile/page
    // Compute the strip of interest, note since TIFF handles
    // full row strips, the x is ignored (for now).
    td = &tiff->tif_dir;
    if (y >= (int)td->td_imagelength || 
      	x >= (int)td->td_imagewidth ||
      	c >= td->td_samplesperpixel) {
	// XXX emit error messege
	return iflStatusEncode(iflBADPARAMS);
    }

    // Compute the tile
    tile = (int)TIFFComputeTile(tiff, x, y, z, c);

    // if not allocated yet just clear it out
    if (td->td_stripbytecount[tile] <=0 || td->td_stripoffset[tile] <= 0) {
	memset(data, 0, pageSizeBytes);
	return iflOKAY;
    }

    // Now that we have the tile, we can offset into the file
    // and start filling in the data block.  The assumption is that
    // the data block has been allocated with the proper amount of
    // memory space.
    if (!TIFFFieldSet(tiff, FIELD_TILEDIMENSIONS) && 
        ((tile+1)%td->td_stripsperimage) == 0 )
       	bytecount = tiff->tif_scanlinesize * (
	    ((tiff->tif_dir.td_compression==COMPRESSION_CCITTFAX3)? size.y-1 : size.y) - 
	    ((int)td->td_rowsperstrip * ((int)td->td_stripsperimage-1)));
    else
	bytecount = pageSizeBytes;

    // Bypass the dumpmode buffer and use the one that's already allocated
    
    beginFileIO();  // XXX shouldn't include decoding as part of file I/O
    if (tiff->tif_dir.td_compression == COMPRESSION_NONE) {
	TIFFReadBufferSetup(tiff, (char *)data, (tsize_t)pageSizeBytes);
	writeInit = FALSE;
    }
 
    int sts = isTiled(tiff)?
	TIFFReadEncodedTile (tiff, tile, (u_char*)data, (tsize_t)bytecount) :
	TIFFReadEncodedStrip(tiff, tile, (u_char*)data, (tsize_t)bytecount);
    endFileIO();
    if (sts == -1)
	return iflStatusEncode(iflREADFAILED, iflSubDomainUNIX, oserror());

    // Since the data is scanline aligned in TIFF, you have to
    // realign it if the byte alignment is not even.
    //
    if ((tiff->tif_dir.td_bitspersample%8) != 0) {
  	bytecount = pageSizeBytes;
  	ifl_tiff_align(tiff, (u_char*)data, (int)bytecount);
    }
    
    return iflOKAY;
}


iflStatus 
iflTIFFFile::setPage(const void* data, int x, int y, int z, int c,
				       int nx, int ny, int nz, int nc)
{
    // Check if the file is writeable
    if (accessmode == O_RDONLY)
	return iflStatusEncode(iflREADONLY);

    // Compute tile.
    int tile = (int)TIFFComputeTile(tiff, x, y, z, c);

    // Compute the bytecount, border tiles may not be full tiles!
    // NOTE: border tiles are assumed to be padded with some value
    // that is also stored in the file but when read, should be ignored.
    TIFFDirectory *td = &tiff->tif_dir;
    int bytecount;
    if (isTiled(tiff))
    	bytecount = pageSize.x * pageSize.y * pageSize.z * pageSize.c * 
		((td->td_bitspersample +7)/8);
    else
    	bytecount = nx * ny * nz * nc * 
		((td->td_bitspersample +7)/8);

    /* If it's bit data, then it has to be realigned */
    if (td->td_bitspersample < 8) {
	int bsize = int((((td->td_tilewidth * td->td_bitspersample 
			* td->td_samplesperpixel)+7)/8) * td->td_tilelength);
  	void* tmpdata = alloca(bsize);
	if (tmpdata == NULL)
	    return iflStatusEncode(iflMALLOCFAILED, iflSubDomainUNIX, oserror());
	bsize = td->td_bitspersample * 
	    ((td->td_planarconfig != PLANARCONFIG_SEPARATE) ? 
	    	td->td_samplesperpixel : 1) * (int)td->td_tilewidth;
	  //((x+td->td_tilewidth >= size.x)? size.x-x:td->td_tilewidth);
	ifl_tiff_bitshift(data, tmpdata, bsize, (int)td->td_tilelength);
	bytecount = tiff->tif_scanlinesize * pageSize.y * pageSize.z;
	data = tmpdata;
    }
  
    beginFileIO();  // XXX shouldn't include encoding as part of file I/O

    // We cannot allow random writes with compressed output for now.
    //
    if (td->td_stripbytecount != NULL && td->td_stripbytecount[tile] > 0) {
    	if (tiff->tif_dir.td_compression == COMPRESSION_NONE) {
    	    // let it just be overwritten with new data (don't append)
    	    td->td_stripbytecount[tile] = 0;
    	    tiff->tif_curoff = 0;
    	}
	else {
	    endFileIO();
	    return iflStatusEncode(iflUNSUPPORTEDBYLIBRARY);
	}
    }

    // If there was a read previously, you need to reset the
    // the buffers
    if (!writeInit) {
	TIFFWriteBufferSetup(tiff, NULL, (tsize_t) -1);
	writeInit = TRUE;
    }

    int sts = isTiled(tiff)?
    	TIFFWriteEncodedTile (tiff, tile, (tdata_t)data, bytecount) :
	TIFFWriteEncodedStrip(tiff, tile, (tdata_t)data, bytecount);
    endFileIO();
    
    if (sts == -1)
	return iflStatusEncode(iflWRITEFAILED, iflSubDomainUNIX, oserror());

    return iflOKAY;
}

iflStatus
iflTIFFFile::setCurrentImg(int i)
{
    if (i == getCurrentImg())
	return iflOKAY;
    if (i < 0 || i >= getNumImgs())
	return iflStatusEncode(iflFILEINDEXOOB);

    if (!TIFFSetDirectory(tiff, i)) {
	iflError(MM_ERROR,
		 "Failed to set current image to %d in TIFF directory", i);
	return iflStatusEncode(iflBADPARAMS);
    } else {
    	TIFFReadBufferSetup(tiff, NULL, 0);
	imgidx = i;
	iflStatus sts = initDir();
	if (sts != iflOKAY) {
	    char buf[400];
	    iflError(MM_ERROR, "Problem reading image in file \"%s\" at index %d: %s\n", getFileName(), i, iflStatusToString(sts, buf, sizeof(buf)));
	    return sts;
	}
	return iflOKAY;
    }
}

iflStatus
iflTIFFFile::getColormap(const iflColormap*& ret)
{
#if 0
    iflAssert(cmap ! NULL);
#endif
    ret = cmap;
    return iflOKAY;
}

iflStatus
iflTIFFFile::setColormap(const iflColormap *newcmap)
{
	if (cm != iflRGBPalette || newcmap == NULL 
	    || newcmap->getNumChans() != 3
	    || (newcmap->getDataType() != iflUChar &&
	    	newcmap->getDataType() != iflUShort))
	    return iflStatusEncode(iflBADPARAMS);

	if (cmap == NULL) cmap = new iflColormap;
	*cmap = *newcmap;

	int len = int(1L<<tiff->tif_dir.td_bitspersample);
	double min, max;
	cmap->getDomain(min, max);
	int alloc = cmap->getDataType() != iflUShort 
		    || min != 0 || max != len-1;
	u_short *data;
	if (alloc) {
	    data = new u_short [len*3];
	    int shift = cmap->getDataType() == iflUChar? 8 : 0;
	    for (int c = 0; c < 3; c++) {
		u_short* chan = data+len*c;
		for (int idx = 0; idx < len; idx++)
		    chan[idx] = int(cmap->getVal(idx, c)) << shift;
	    }
	}
	else 
	    data = (u_short*)cmap->getData();
	if (data == NULL || !TIFFSetField(tiff, TIFFTAG_COLORMAP, 
					  data, data+len, data+len*2))
	    return iflStatusEncode(iflBADFIELDSET);
	if (alloc) delete data;

	return iflOKAY;
}

iflStatus
iflTIFFFile::setStatMinMax(double min, double max)
{
    // Only use the
    // new min/max tags if the data type is beyond the scope of the old tags.

	// XXX assumes values are powers of 2
	if (dtype & ~(iflBit|iflUChar|iflUShort)) {
	    if (!TIFFSetField(tiff, TIFFTAG_SMINSAMPLEVALUE, min))
		return iflStatusEncode(iflBADFIELDSET);
	}
	else {
	    if (!TIFFSetField(tiff, TIFFTAG_MINSAMPLEVALUE, int(min)))
		return iflStatusEncode(iflBADFIELDSET);
	}

	if (dtype & ~(iflBit|iflUChar|iflUShort)) {
	    if (!TIFFSetField(tiff, TIFFTAG_SMAXSAMPLEVALUE, max))
		return iflStatusEncode(iflBADFIELDSET);
	}
	else {
	    if (!TIFFSetField(tiff, TIFFTAG_MAXSAMPLEVALUE, int(max)))
		return iflStatusEncode(iflBADFIELDSET);
	}
	return iflOKAY;
}


iflStatus
iflTIFFFile::getItem(int tag, va_list ap)
{
    int ret;
    iflAssert(tiff != NULL);

    ret = TIFFVGetField(tiff, tag, ap);
    
    return ret ? iflOKAY : iflStatusEncode(iflBADFIELDGET);
}


iflStatus
iflTIFFFile::setItem(int tag, va_list ap)
{
    int ret;
    iflAssert(tiff != NULL);

    // Don't allow the image size attributes to change

    if (tiff->tif_mode == O_RDONLY) return iflStatusEncode(iflREADONLY);
	    
    if (tag == TIFFTAG_IMAGEWIDTH    || tag == TIFFTAG_IMAGELENGTH ||
	tag == TIFFTAG_BITSPERSAMPLE || tag == TIFFTAG_SAMPLESPERPIXEL || 
	tag == TIFFTAG_ROWSPERSTRIP  || tag == TIFFTAG_TILEWIDTH ||
	tag == TIFFTAG_TILELENGTH    || tag == TIFFTAG_TILEDEPTH ||
	tag == TIFFTAG_DATATYPE      || tag == TIFFTAG_IMAGEDEPTH ||
	tag == TIFFTAG_SAMPLEFORMAT  ||
	tag == TIFFTAG_PLANARCONFIG  || tag == TIFFTAG_COLORMAP)
	return iflStatusEncode(iflUNSUPPORTEDBYFORMAT);

    ret = TIFFVSetField(tiff, tag, ap);
    
    return ret? iflOKAY:iflStatusEncode(iflUNSUPPORTEDBYFORMAT);
}

iflStatus
iflTIFFFile::flush()
{
    if (tiff != NULL)
	TIFFFlushData(tiff);
    return iflOKAY;
}


/*
 * Append to the file a new image (directory)
 */
iflStatus
iflTIFFFile::appendImg(iflFile* source, iflFileConfig* cfg)
{
    iflFileConfig fc;
    if (cfg) fc = *cfg;
    iflStatus sts = format->applyDefaults(source, fc);
    if (sts != iflOKAY) return sts;

    // flush the current image and write out the directory
    TIFFWriteDirectory(tiff);

    // force to the last sub-image
    setCurrentImg(nimgs-1);

    iflSize size, pSize;
    fc.getSize(size);
    fc.getPageSize(pSize);

    sts = initImg(size, fc.getDataType(), fc.getOrder(), fc.getOrientation(), 
		  pSize, fc.getCompression(), fc.getColorModel());
    if (sts != iflOKAY) {
	char buf[400];
	iflError(MM_ERROR, "Couldn't append image to \"%s\" at index %d: %s\n", 
		 getFileName(), imgidx, 
		 iflStatusToString(sts, buf, sizeof(buf)));
	return sts;
    }

    nimgs++;
    imgidx++;

    // forget any old colormap
    delete cmap; 
    cmap = NULL; 

    return iflOKAY;
}


//
// repack the bits from scanline/#bits alignment to a tile alignment
//
static int
ifl_tiff_align(TIFF* tif, u_char* buf, int cc)
{
    TIFFDirectory *td = &tif->tif_dir;
    int nbits, i, j, k;
    u_char *ibuf, *tmp;

    // Since the data is packed by scanlines, we have to
    // repack it to be as tiles.
    //
    int cnt = int(td->td_tilewidth * 
	      ((td->td_planarconfig == PLANARCONFIG_SEPARATE) ? 1 :
	      td->td_samplesperpixel));
    nbits = td->td_bitspersample * cnt;
    int len = (nbits+7) >> 3;
    int iter = int((isTiled(tif))
	? tif->tif_dir.td_tilelength
	: (tif->tif_dir.td_rowsperstrip == 0xffffffffL
	   ? tif->tif_dir.td_imagelength
	   : tif->tif_dir.td_rowsperstrip));

    // If the data is aligned by row and not tile, repack it 
    //
    if ((nbits%8) != 0) {
	ibuf = buf + len;
	tmp = ibuf - 1;
	j = nbits;
	for (i=2; i<=iter; i++) {
	    k = 7 - (j % 8);
	    j += nbits;
	    if (i%8 == 1) 
		memcpy(tmp, ibuf, len);
	    else 
		ifl_tiff_bitcopy(ibuf, tmp, k, nbits);
	    tmp = buf + (j>>3);
	    ibuf += len;
	}
    }

    // Unpack the bits if possible.  Assume that enough memory was 
    // allocated to unpack them into bytes.
    //
    if (tif->tif_dir.td_bitspersample > 1 &&
	tif->tif_dir.td_bitspersample < 8) 
	ifl_tiff_bitstobytes(buf, buf, tif->tif_dir.td_bitspersample, cc);

    return (0);
}


//
// Shift a stream of bits right
//
static int rmsk[9] = {0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};
static int lmsk[9] = {0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff};

static void
ifl_tiff_bitshift(const void* ibuf, void* obuf, int nbits, int nlines)
{
    const u_char *from;
    u_char *to;
    int i, j, k, l, m, x, y;

    i = nbits % 8;		// bits in last byte
    if (i == 0) { 		
	// if the bits are perfectly aligned, do a straight copy
	memcpy(obuf,ibuf,(nbits>>3)*nlines);
	return;
    }

    j = nbits >> 3;		// number of bytes/line - 1
    k = nbits*nlines;
    l = k % 8;		// last packed bits
    k = k >> 3;

    // always shift right
    from = &((const u_char*)ibuf)[k];
    to = &((u_char*)obuf)[((j+1)*nlines) - 1];

    // get the last bits in place
    for (x = 0; x<nlines; x++) {
	*to = (*from & rmsk[l]); 
	if (l==i) {
	    // copy remaining bytes
	    if (obuf < to) {
		to -= j;
		from -= j;
		memcpy(to, from, j);
		to--;
		l = 0;
	    }
	    else
		return;
	}
	else {
	    if (l > i) {
		l = l - i;
		*to-- = *to << l;
		m = 8 - l;
	    }
	    else {	
		l = i-l;
		*to-- = (*to >> l) | ((*(from-1) & lmsk[l]) << (8-l));
		l = 8 - l;
		m = 8 - l;
		from--;
	    }
	    for (y=0; y<j; y++) { 
		*to-- = ((*from & rmsk[l]) >> m) | 
		    ((*(from-1) & lmsk[m]) << l); 
		from--;
	    }
	}
    }
}	


// Copy the non-byte aligned bits from one stream to another.
// Do it one bit at a time for a length of nbits.
//
static void
ifl_tiff_bitcopy(const void* ibuf, void* obuf, int startbit, int nbits)
{
    const u_char* from = (const u_char*)ibuf;
    u_char* to = (u_char*)obuf;
    int i, j, k, l;

    j = 7 - startbit;
    k = startbit+1;
    l = nbits >> 3;
    if (startbit == 7) {
	memcpy(to, from, l);
	k = nbits % 8;
	to[l] = from[l]&rmsk[k];
    }
    else {
	to[0] = (to[0]&rmsk[j]) | ((from[0]&rmsk[k])>>j);

	if (nbits < 8) {
	    if (nbits != k)
		to[1] = from[0] << k;
	    return;
	}

	for(i=1; i<=l; i++) {
	    to[i] = ((from[i-1]&lmsk[j])<< k) | 
		((from[i]&rmsk[k+1])>>j);
	}

	l = (nbits-k) % 8;
	if (l != 0)  		// there's one more byte to do!
	    to[i] = from[i-1]&rmsk[l];
	else
	    to[i] = ((from[i-2]&lmsk[j])<< k) | 
		((from[i-1]&rmsk[k+1])>>j);
    }
}


// Unpack samples that are < 8 bits into bytes:
//
// char *from : packed bits
// char *to   : resulting unpacked bits
// int   bs   : number of bits per sample
// int   cnt  : number of samples
// 
static void
ifl_tiff_bitstobytes(const void* ibuf, void* obuf, int bs, int cnt)
{
    const u_char* from = (const u_char*)ibuf;
    u_char* to = (u_char*)obuf;
    int totbits, packedcnt;
    register int i, j, bitpos;

    totbits = bs * cnt; 		// total number of bits used
    packedcnt = (totbits / 8);		// packed buffer size minus 1
    bitpos = totbits % 8;		// last bit position
    if (bitpos != 0)
	bitpos = 8 - bitpos;

    j = (bitpos==0) ? packedcnt-1 : packedcnt;
    for (i=cnt-1; i>=0; i--) {
	if (bitpos == 0) {
	    to[i] = from[j] & lmsk[bs];
	    bitpos = bs;
	}
	else if (bitpos+bs < 8) {
	    to[i] = (from[j] >> bs) & lmsk[bs];
	    bitpos += bs;
	}
	else {
	    to[i] = (from[j] >> bitpos) | (from[j-1] << (8-bitpos)) & lmsk[bs];
	    j--;
	    bitpos = (bitpos + bs) % 8;
	}
    }
}
iflStatus
iflTIFFFile::_getICCProfile(int& size, void*& profile)
{
    int ret;
    iflAssert(tiff != NULL);

    ret = TIFFGetField(tiff, TIFFTAG_ICCPROFILE, &size, &profile);

    return ret ? iflOKAY : iflStatusEncode(iflBADFIELDGET);
}


iflStatus
iflTIFFFile::_freeICCProfile(void*)
{
    // don't do anything, we manage the memory internally
    return iflOKAY;
}


iflStatus
iflTIFFFile::_setICCProfile(int size, const void* profile)
{
    int ret;
    iflAssert(tiff != NULL);

    ret = TIFFSetField(tiff, TIFFTAG_ICCPROFILE, size, profile);

    return ret? iflOKAY:iflStatusEncode(iflBADFIELDSET);
}

