#include <cstdio>

#include "error.h"
#include "glutwm.h"
#include "window.h"

#define MAXWINDOWS 256

// initial window size and mode
int width = 600, height = 600;
int mode = GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE;

// window pointers
static Window *window[MAXWINDOWS];
// set of windows that have an enabled idle method
static int idle[MAXWINDOWS];
static int nidle = 0;
// external user idle callback
void (*useridle)(void) = NULL;

// functions to set glut callbacks
static void (GLUTWMCB *setidle)(void (*f)(void)) = glutIdleFunc;
static void (GLUTWMCB *setdisplay)(void (*f)(void)) = glutDisplayFunc;
static void (GLUTWMCB *setmouse)(void (*f)(int,int,int,int)) = glutMouseFunc;
static void (GLUTWMCB *setmotion)(void (*f)(int,int)) = glutMotionFunc;
static void (GLUTWMCB *setkeyboard)(void (*f)(unsigned char,int,int)) = glutKeyboardFunc;
static void (GLUTWMCB *setreshape)(void (*f)(int, int)) = glutReshapeFunc;

// set control functions
static void insert(int id)
{
    for (int i = 0; i < nidle; i++) if (idle[i] == id) return;
    if (nidle < MAXWINDOWS) idle[nidle++] = id;
}

static void remove(int id)
{
    int i;
    for (i = 0; i < nidle; i++) if (idle[i] == id) break;
    if (i >= nidle) return;
    for (int j = i; j < nidle-1; j++) idle[j] = idle[j+1];
    nidle--;
}

// callbacks
void idle_cb(void)
{
    for (int i = 0; i < nidle; i++) {
        glutSetWindow(idle[i]);
        window[idle[i]]->Idle();
    }
    if (useridle) useridle();
}

static void display_cb(void)
{
    int id = glutGetWindow();
    if (id < MAXWINDOWS && window[id]) window[id]->Display();
    glutSwapBuffers();
}

static void reshape_cb(int w, int h)
{
    int id = glutGetWindow();
    if (id < MAXWINDOWS && window[id]) window[id]->Reshape(w, h);
}

static void keyboard_cb(unsigned char c, int x, int y)
{
    int id = glutGetWindow();
    if (id < MAXWINDOWS && window[id]) {
        window[id]->Adjust(&x, &y);
        window[id]->Keyboard(c, x, y);
    }
}

static void motion_cb(int x, int y)
{
    int id = glutGetWindow();
    if (id < MAXWINDOWS && window[id]) {
        window[id]->Adjust(&x, &y);
        window[id]->Motion(x, y);
    }
}

static void mouse_cb(int button, int state, int x, int y)
{
    int id = glutGetWindow();
    if (id < MAXWINDOWS && window[id]) {
        window[id]->Adjust(&x, &y);
        window[id]->Button(button, state, x, y);
    }
}

// API functions
void GlutWMOpen(int *argc, char *argv[])
{
    glutInit(argc, argv); 
    for (int i = 0; i < MAXWINDOWS; i++) {
        window[i] = NULL;
        idle[i] = -1;
        nidle = 0;
    }
}

void GlutWMIdleFunc(void (*func)(void))
{
    useridle = func;
    if (!useridle && nidle == 0) setidle(NULL);
    if (useridle) setidle(idle_cb);
}

void GlutWMStartIdle(int id)
{
    if (id < 0) id = glutGetWindow();
    if (id > 0 && id < MAXWINDOWS) {
        insert(id);
        setidle(idle_cb);
    }
}

void GlutWMStopIdle(int id)
{
    if (id < 0) id = glutGetWindow();
    if (id > 0 && id < MAXWINDOWS) {
        remove(id);
        if (nidle == 0 && !useridle) setidle(NULL);
    }
}

void GlutWMClose(void) 
{ 
    setdisplay(NULL);
    setmouse(NULL);
    setkeyboard(NULL);
    setreshape(NULL);
    setmotion(NULL);
    setidle(NULL);
}

void GlutWMPostRedisplay(int id)
{
    if (id > 0) {
        int old_id = glutGetWindow();
        glutSetWindow(id);
        glutPostRedisplay();
        if (old_id > 0) glutSetWindow(old_id);
    } else glutPostRedisplay();
}


void GlutWMShowWindow(int id)
{
    if (id > 0) {
        int old_id = glutGetWindow();
        glutSetWindow(id);
        glutShowWindow();
        if (old_id > 0) glutSetWindow(old_id);
    } else glutShowWindow();
}

void GlutWMHideWindow(int id)
{
    if (id > 0) {
        int old_id = glutGetWindow();
        glutSetWindow(id);
        glutHideWindow();
        if (old_id > 0) glutSetWindow(old_id);
    } else glutHideWindow();
}


void GlutWMSwapBuffers(int id)
{
    if (id > 0) {
        int old_id = glutGetWindow();
        glutSetWindow(id);
        glutSwapBuffers();
        if (old_id > 0) glutSetWindow(old_id);
    } else glutSwapBuffers();
}

void GlutWMFullScreen(int id)
{
    if (id > 0) {
        int old_id = glutGetWindow();
        glutSetWindow(id);
        glutFullScreen();
        if (old_id > 0) glutSetWindow(old_id);
    } else glutFullScreen();
}

int GlutWMDestroyWindow(int id)
{
    debugf(("Destroying window %d", id));
    if (id > 0 && id < MAXWINDOWS && window[id]) {
        glutDestroyWindow(id);
        window[id]->Destroy();
        return 1;
    }
    return 0;
}

void GlutWMInitWindowSize(int w, int h)
{
    width = w; height = h;
}

void GlutWMInitDisplayMode(int m)
{
    mode  = m;
}

int GlutWMReplaceWindow(int id, const char *name, Window *wnd)
{
    if (id < 1 || id >= MAXWINDOWS) return 0;
    window[id] = wnd;
    int old_id = glutGetWindow();
    glutSetWindow(id);
    glutSetWindowTitle(name);
    window[id]->Init(id);
    if (old_id > 0) glutSetWindow(old_id);
    return 1;
}

int GlutWMCreateWindow(const char *name, Window *wnd)
{
    if (wnd->GetId() != 0) {
        errorf(("Window object already in use!"));
        return 0;
    }
    glutInitWindowSize(width, height);
    glutInitDisplayMode(mode);
    int id = glutCreateWindow(name);
    if (id >= MAXWINDOWS || id < 1) {
        if (id >= MAXWINDOWS) {
            errorf(("Too many windows!"));
            glutDestroyWindow(id);
        } else errorf(("glutCreateWindow failed!"));
        return 0;
    }
    debugf(("Created window %d", id));
    window[id] = wnd;
    setdisplay(display_cb);
    setmouse(mouse_cb);
    setkeyboard(keyboard_cb);
    setreshape(reshape_cb);
    setmotion(motion_cb);
    window[id]->Init(id);
    glutPostRedisplay();
    return id;
}

void GlutWMPositionWindow(int id, int x0, int y0)
{
    int old_id = glutGetWindow();
    glutSetWindow(id);
    glutPositionWindow(x0, y0);
    if (old_id > 0) glutSetWindow(old_id);
}

void GlutWMReshapeWindow(int id, int w, int h)
{
    int old_id = glutGetWindow();
    glutSetWindow(id);
    glutReshapeWindow(w, h);
    if (old_id > 0) glutSetWindow(old_id);
}

void GlutWMMainLoop(void)
{ glutMainLoop(); }

// Meta callback functions
void GlutWMMetaIdleFunc(void (GLUTWMCB *set)(void (*f)(void)))
{ setidle = set; }
void GlutWMMetaDisplayFunc(void (GLUTWMCB *set)(void (*f)(void)))
{ setdisplay = set; }
void GlutWMMetaMouseFunc(void (GLUTWMCB *set)(void (*f)(int, int, int ,int)))
{ setmouse = set; }
void GlutWMMetaMotionFunc(void (GLUTWMCB *set)(void (*f)(int, int)))
{ setmotion = set; }
void GlutWMMetaReshapeFunc(void (GLUTWMCB *set)(void (*f)(int, int)))
{ setreshape = set; }
void GlutWMMetaKeyboardFunc(void (GLUTWMCB *set)(void (*f) (unsigned char, int, int)))
{ setkeyboard = set; }

// Buffer drawing helpers
void GlutWMDrawPixels(GLsizei width, GLsizei height, GLenum format,
        const unsigned char *data)
{ glDrawPixels(width, height, format, GL_UNSIGNED_BYTE, data); }
void GlutWMDrawPixels(GLsizei width, GLsizei height, GLenum format,
        const unsigned short *data)
{ glDrawPixels(width, height, format, GL_UNSIGNED_SHORT, data); }
void GlutWMDrawPixels(GLsizei width, GLsizei height, GLenum format,
        const unsigned long *data)
{ glDrawPixels(width, height, format, GL_UNSIGNED_INT, data); }
void GlutWMDrawPixels(GLsizei width, GLsizei height, GLenum format,
        const float *data)
{ glDrawPixels(width, height, format, GL_FLOAT, data); }


// Buffer reading helpers
void GlutWMReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, 
        GLenum format, unsigned char *data)
{ glReadPixels(x, y, width, height, format, GL_UNSIGNED_BYTE, data); }
void GlutWMReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, 
        GLenum format, unsigned short *data)
{ glReadPixels(x, y, width, height, format, GL_UNSIGNED_SHORT, data); }
void GlutWMReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, 
        GLenum format, unsigned long *data)
{ glReadPixels(x, y, width, height, format, GL_UNSIGNED_INT, data); }
void GlutWMReadPixels(GLint x, GLint y, GLsizei width, GLsizei height, 
        GLenum format, float *data)
{ glReadPixels(x, y, width, height, format, GL_FLOAT, data); }
