#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <stdlib.h>

#include "rply.h"

/* ----------------------------------------------------------------------
 * Constants 
 * ---------------------------------------------------------------------- */
#define WORDSIZE 256
#define BUFFERSIZE (8*1024)

typedef enum e_ply_mode_ {
    PLY_READ, 
    PLY_WRITE
} e_ply_mode;

static const char *const ply_format_list[] = {
    "binary_big_endian", "binary_little_endian", "ascii", NULL
};     /* order matches e_ply_format enum */

static const char *const ply_type_list[] = {
    "int8", "uint8", "int16", "uint16", 
    "int32", "uint32", "float32", "float64",
    "char", "uchar", "short", "ushort", 
    "int", "uint", "float", "double",
    "list", NULL
};     /* order matches e_ply_type enum */

/* ----------------------------------------------------------------------
 * Property information
 *
 * name: name of this property
 * type: type of this property (list or type of scalar value)
 * type_nentries, type_value: type of list property count and values
 * callback: function to be called when this property is called
 *
 * Returns 1 if should continue processing file, 0 if should abort.
 * ---------------------------------------------------------------------- */
typedef struct t_ply_property_ {
    char name[WORDSIZE];
    int type;
    int value_type;
    int nentries_type;
    p_ply_cb callback;
    void *pdata;
    long idata;
} t_ply_property; 
typedef t_ply_property *p_ply_property;

/* ----------------------------------------------------------------------
 * Element information
 *
 * name: name of this property
 * nentries: number of elements of this type in file
 * property: property descriptions for this element
 * nproperty: number of properties in this element
 *
 * Returns 1 if should continue processing file, 0 if should abort.
 * ---------------------------------------------------------------------- */
typedef struct t_ply_element_ {
    char name[WORDSIZE];
    long nelements;
    p_ply_property property;
    int nproperties;
} t_ply_element;
typedef t_ply_element *p_ply_element;

/* ----------------------------------------------------------------------
 * Input/output driver
 *
 * Depending on file format, different functions are used to read/write 
 * property fields. The drivers make it transparent to read/write in ascii, 
 * big endian or little endian cases.
 * ---------------------------------------------------------------------- */
typedef int (*p_ply_ihandler)(p_ply ply, double *value);
typedef int (*p_ply_ichunk)(p_ply ply, void *anydata, size_t size);
typedef struct t_ply_idriver_ {
    p_ply_ihandler ihandler[16];
    p_ply_ichunk ichunk;
    const char *name;
} t_ply_idriver;
typedef t_ply_idriver *p_ply_idriver;

typedef int (*p_ply_ohandler)(p_ply ply, double value);
typedef int (*p_ply_ochunk)(p_ply ply, void *anydata, size_t size);
typedef struct t_ply_odriver_ {
    p_ply_ohandler ohandler[16];
    p_ply_ochunk ochunk;
    const char *name;
} t_ply_odriver;
typedef t_ply_odriver *p_ply_odriver;

/* ----------------------------------------------------------------------
 * Ply file handle. 
 *
 * mode: read or write (from e_ply_mode)
 * file: format of file associated with handle (from e_ply_format)
 * element: elements description for this file
 * nelement: number of different elements in file
 * fp: file pointer associated with ply file
 * c: last character read from ply file
 * buffer: last word/chunck of data read from ply file
 * line_number, column_number: last line/column processed in file
 * idriver, odriver: input driver used to get property fields from file 
 * argument: storage space for callback arguments
 * current_element, current_property: element and property being written
 * current_element_entry, current_property_entry: current entry bein written
 * ---------------------------------------------------------------------- */
typedef struct t_ply_ {
    int mode, format;
    p_ply_element element;
    int nelements;
    FILE *fp;
    int c;
    int line_number, column_number;
    char buffer[BUFFERSIZE];
    int buffer_first, buffer_last;
    p_ply_idriver idriver;
    p_ply_odriver odriver;
    t_ply_argument argument;
    long welement, wproperty;
    long welement_index, wentry_index, wnentries;
} t_ply;

/* ----------------------------------------------------------------------
 * Constants 
 * ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 * I/O stuff
 * ---------------------------------------------------------------------- */
static t_ply_idriver ply_idriver_ascii;
static t_ply_idriver ply_idriver_binary;
static t_ply_idriver ply_idriver_binary_reverse;
static t_ply_odriver ply_odriver_ascii;
static t_ply_odriver ply_odriver_binary;
static t_ply_odriver ply_odriver_binary_reverse;

static int ply_read_word(p_ply ply);
static int ply_read_chunk(p_ply ply, void *anybuffer, size_t size);
static int ply_read_chunk_reverse(p_ply ply, void *anybuffer, size_t size);
static int ply_write_chunk(p_ply ply, void *anybuffer, size_t size);
static int ply_write_chunk_reverse(p_ply ply, void *anybuffer, size_t size);
static void ply_reverse(void *anydata, size_t size);

/* ----------------------------------------------------------------------
 * String functions
 * ---------------------------------------------------------------------- */
static int ply_find_string(const char *item, const char* const list[]);
static p_ply_element ply_find_element(p_ply ply, const char *name);
static p_ply_property ply_find_property(p_ply_element element,const char *name);

/* ----------------------------------------------------------------------
 * Header parsing
 * ---------------------------------------------------------------------- */
static int ply_read_format(p_ply ply);
static int ply_read_comment(p_ply ply);
static int ply_read_property(p_ply ply);
static int ply_read_element(p_ply ply);

/* ----------------------------------------------------------------------
 * Memory allocation and initialization
 * ---------------------------------------------------------------------- */
static void ply_init(p_ply ply);
static void ply_element_init(p_ply_element element);
static void ply_property_init(p_ply_property property);
static p_ply ply_alloc(void);
static p_ply_element ply_grow_elements(p_ply ply);
static p_ply_property ply_grow_property(p_ply_element element);

/* ----------------------------------------------------------------------
 * Special functions
 * ---------------------------------------------------------------------- */
static int ply_arch_endian(void);
static int ply_type_check(void); 
static void ply_print(p_ply ply);
static void ply_error(const char *fmt, ...);

#define errorf(args) do { \
    ply_error("error: %s: %d: ", __FILE__, __LINE__); \
        ply_error args; \
        ply_error("\n"); \
} while(0)

/* ----------------------------------------------------------------------
 * Exported functions
 * ---------------------------------------------------------------------- */
/* ----------------------------------------------------------------------
 * Read support functions
 * ---------------------------------------------------------------------- */
p_ply ply_open(const char *name)
{
    char magic[5] = "    ";
    FILE *fp = NULL; 
    p_ply ply = NULL;
    if (!ply_type_check()) {
        errorf(("Incompatible type system"));
        return NULL;
    }
    assert(name);
    if (!name) return NULL;
    fp = fopen(name, "rb");
    if (!fp) {
        errorf(("Unable to open file", name));
        return NULL;
    }
    if (fread(magic, 1, 4, fp) < 4) {
        errorf(("Error reading from file"));
        fclose(fp);
        return NULL;
    }
    if (strcmp(magic, "ply\n")) {
        errorf(("Expected magic number 'ply\\n'"));
        fclose(fp);
        return NULL;
    }
    ply = ply_alloc();
    if (!ply) return NULL;
    ply->fp = fp;
    ply->mode = PLY_READ;
    return ply;
}

int ply_read_header(p_ply ply)
{
    assert(ply && ply->fp && ply->mode == PLY_READ);
    if (!ply || !ply->fp || ply->mode != PLY_READ) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    if (!ply_read_word(ply)) return 0;
    /* parse file format */
    if (!ply_read_format(ply)) {
        errorf(("Invalid format at line %d", ply->line_number));
        return 0;
   }
    /* then parse elements or comments until the end of header */
    while (strcmp(ply->buffer, "end_header")) {
       if (!ply_read_comment(ply) && !ply_read_element(ply)) {
            errorf(("Expected comment, element or end_header at line %d",
                        ply->line_number));
            return 0;
        }
    }
    return 1;
}

int ply_set_read_cb(p_ply ply, const char *element_name, 
        const char* property_name, p_ply_cb callback, 
        void *pdata, long idata)
{
    p_ply_element element = NULL; 
    p_ply_property property = NULL;
    assert(ply && element_name && property_name);
    if (!ply || !element_name || !property_name) {
        errorf(("Invalid arguments"));
        return 0;
    }
    element = ply_find_element(ply, element_name);
    if (!element) return 0;
    property = ply_find_property(element, property_name);
    if (!property) return 0;
    property->callback = callback;
    property->pdata = pdata;
    property->idata = idata;
    return element->nelements;
}

int ply_read(p_ply ply)
{
    long i, j, k, l;
    p_ply_argument argument;
	p_ply_ihandler *driver = NULL;
	p_ply_ihandler handler = NULL;
    assert(ply && ply->fp && ply->mode == PLY_READ);
    if (!ply || !ply->fp || ply->mode != PLY_READ) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    argument = &ply->argument;
	driver = ply->idriver->ihandler; 
    /* for each element type */
    for (i = 0; i < ply->nelements; i++) {
        p_ply_element element = &ply->element[i];
        argument->element = element->name;
        argument->nelements = element->nelements;
        /* for each element of this type */
        for (j = 0; j < element->nelements; j++) {
            argument->element_index = j;
            for (k = 0; k < element->nproperties; k++) {
                p_ply_property property = &element->property[k];
                p_ply_cb callback = property->callback;
                argument->property = property->name;
                argument->pdata = property->pdata;
                argument->idata = property->idata;
                if (property->type == PLY_LIST) {
                    double nentries;
                    handler = ply->idriver->ihandler[property->nentries_type];
                    if (!handler(ply, &nentries)) return 0;
                    handler = driver[property->value_type];
                    argument->nentries = (long) nentries;
                    for (l = 0; l < (long) nentries; l++) {
                        argument->entry_index = l;
                        if (!handler(ply, &argument->value)) return 0;
                        if (callback && !callback(argument)) return 0;
                    }
                } else {
                    argument->nentries = 1;
                    argument->entry_index = 0;
                    handler = driver[property->type];
                    if (!handler(ply, &argument->value)) return 0;
                    if (callback && !callback(argument)) return 0;
                }
            }
        }
    }
    return 1;
}

/* ----------------------------------------------------------------------
 * Write support functions
 * ---------------------------------------------------------------------- */
p_ply ply_create(const char *name, int format)
{
    FILE *fp = NULL;
    p_ply ply = NULL;
    if (!ply_type_check()) {
        errorf(("Incompatible type system"));
        return NULL;
    }
    assert(name && format >= 0 && format <= PLY_DEFAULT);
    if (!name || format < 0 || format > PLY_DEFAULT) {
        errorf(("Invalid arguments"));
        return NULL;
    }
    fp = fopen(name, "wb");
    if (!fp) {
        errorf(("Unable to create file '%s'", name));
        return NULL;
    }
    ply = ply_alloc();
    if (!ply) return NULL;
    ply->mode = PLY_WRITE;
    if (format == PLY_DEFAULT) format = ply_arch_endian();
    if (format == PLY_ASCII) ply->odriver = &ply_odriver_ascii;
    else if (format == ply_arch_endian()) ply->odriver = &ply_odriver_binary;
    else ply->odriver = &ply_odriver_binary_reverse;
    ply->format = format;
    ply->fp = fp;
    return ply;
}

int ply_add_element(p_ply ply, const char *name, long nelements)
{
    p_ply_element element = NULL;
    assert(ply && ply->fp && ply->mode == PLY_WRITE);
    if (!ply || !ply->fp || ply->mode != PLY_WRITE) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    assert(name && strlen(name) < 256 && nelements >= 0);
    if (!name || strlen(name) >= 256 || nelements < 0) {
        errorf(("Invalid arguments"));
        return 0;
    }
    element = ply_grow_elements(ply);
    if (!element) return 0;
    strcpy(element->name, name);
    element->nelements = nelements;
    return 1;
}

int ply_add_property(p_ply ply, const char *name, int type, int nentries_type,
                int value_type) {
    p_ply_element element = NULL;
    p_ply_property property = NULL;
    assert(ply && ply->fp && ply->mode == PLY_WRITE);
    if (!ply || !ply->fp || ply->mode != PLY_WRITE) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    assert(name && strlen(name) < 256 >= 0);
    if (!name || strlen(name) >= 256) {
        errorf(("Invalid arguments"));
        return 0;
    }
    assert(type >= 0 && type <= PLY_LIST);
    assert(nentries_type >= 0 && nentries_type < PLY_LIST);
    assert(value_type >= 0 && value_type < PLY_LIST);
    if (type < 0 || type > PLY_LIST || 
            nentries_type < 0 || nentries_type > PLY_LIST-1 ||
            value_type < 0 || value_type > PLY_LIST-1) {
        errorf(("Invalid arguments"));
        return 0;
    }
    element = &ply->element[ply->nelements-1];
    property = ply_grow_property(element);
    if (!property) return 0;
    strcpy(property->name, name);
    property->type = type;
    property->nentries_type = nentries_type;
    property->value_type = value_type;
    return 1;
}

int ply_write_header(p_ply ply)
{
    int i, j;
    assert(ply && ply->fp && ply->mode == PLY_WRITE);
    if (!ply || !ply->fp || ply->mode != PLY_WRITE) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    assert(ply->element || ply->nelements == 0); 
    assert(!ply->element || ply->nelements > 0); 
    if (fprintf(ply->fp, "ply\nformat %s 1.0\n", 
                ply_format_list[ply->format]) <= 0) goto error;
    for (i = 0; i < ply->nelements; i++) {
        p_ply_element element = &ply->element[i];
        assert(element->property || element->nproperties == 0); 
        assert(!element->property || element->nproperties > 0); 
        if (fprintf(ply->fp, "element %s %ld\n", element->name, 
                    element->nelements) <= 0) goto error;
        for (j = 0; j < element->nproperties; j++) {
            p_ply_property property = &element->property[j];
            if (property->type == PLY_LIST) {
                if (fprintf(ply->fp, "property list %s %s %s\n", 
                            ply_type_list[property->nentries_type],
                            ply_type_list[property->value_type],
                            property->name) <= 0) goto error;
            } else {
                if (fprintf(ply->fp, "property %s %s\n", 
                            ply_type_list[property->type],
                            property->name) <= 0) goto error;
            }
        }
    }
    return fprintf(ply->fp, "end_header\n") > 0;
error:
    errorf(("Error writing to file"));
    return 0;
}

int ply_write(p_ply ply, double value)
{
    p_ply_element element = NULL;
    p_ply_property property = NULL;
    int type = -1;
    int breakafter = 0;
    if (ply->welement > ply->nelements) return 0;
    element = &ply->element[ply->welement];
    if (ply->wproperty > element->nproperties) return 0;
    property = &element->property[ply->wproperty];
    if (property->type == PLY_LIST) {
        if (ply->wentry_index == 0) {
            type = property->nentries_type;
            ply->wnentries = (long) value;
        } else type = property->value_type;
    } else type = property->type;
    if (!ply->odriver->ohandler[type](ply, value)) {
        errorf(("Failed writing %s of %s %d", 
                    property->name, element->name, ply->welement_index));
        return 0;
    }
    ply->wentry_index++;
    if (ply->wentry_index > ply->wnentries) {
        ply->wentry_index = 0;
        ply->wproperty++;
    }
    if (ply->wproperty >= element->nproperties) {
        ply->wproperty = 0;
        ply->welement_index++;
        if (ply->format == PLY_ASCII) breakafter = 1;
    }
    if (ply->welement_index >= element->nelements) {
        ply->welement_index = 0;
        ply->welement++;
    }
    return !breakafter || putc('\n', ply->fp) > 0;
}

int ply_close(p_ply ply)
{
    int i;
    assert(ply && ply->fp);
    if (!ply || !ply->fp) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    assert(ply->element || ply->nelements == 0);
    assert(!ply->element || ply->nelements > 0);
    if (!ply->element && ply->nelements != 0) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    if (ply->element && ply->nelements <= 0) {
        errorf(("Invalid ply handle"));
        return 0;
    }
    /* write last chunk to file */
    if (ply->mode == PLY_WRITE && 
      fwrite(ply->buffer, 1, ply->buffer_last, ply->fp) < ply->buffer_last) {
        errorf(("Error closing up"));
        return 0;
    }
    fclose(ply->fp);
    /* free all memory used by handle */
    if (ply->element) {
        for (i = 0; i < ply->nelements; i++) {
            p_ply_element element = &ply->element[i];
            if (element->property) free(element->property);
        }
        free(ply->element);
    }
    free(ply);
    return 1;
}

/* ----------------------------------------------------------------------
 * Internal functions
 * ---------------------------------------------------------------------- */
static int ply_find_string(const char *item, const char* const list[])
{
    int i;
    assert(item && list);
    for (i = 0; list[i]; i++) 
        if (!strcmp(list[i], item)) return i;
    return -1;
}

static p_ply_element ply_find_element(p_ply ply, const char *name)
{
    p_ply_element element;
    int i, nelements;
    assert(ply && name); 
    element = ply->element;
    nelements = ply->nelements;
    assert(element || nelements == 0); 
    assert(!element || nelements > 0); 
    for (i = 0; i < nelements; i++) 
        if (!strcmp(element[i].name, name)) return &element[i];
    return NULL;
}

static p_ply_property ply_find_property(p_ply_element element, const char *name)
{
    p_ply_property property;
    int i, nproperties;
    assert(element && name); 
    property = element->property;
    nproperties = element->nproperties;
    assert(property || nproperties == 0); 
    assert(!property || nproperties > 0); 
    for (i = 0; i < nproperties; i++) 
        if (!strcmp(property[i].name, name)) return &property[i];
    return NULL;
}

static int ply_read_word(p_ply ply)
{
    int i = 0;
    assert(ply && ply->fp && ply->mode == PLY_READ);
    /* skip white space */
    while (isspace(ply->c)) {
        if (ply->c == '\n') {
            ply->line_number++;
            ply->column_number = 1;
        }
        ply->c = fgetc(ply->fp);
        ply->column_number++;
    }
    /* get word until it ends, the file ends, or our buffer ends */
    while (!isspace(ply->c) && ply->c != EOF && i < WORDSIZE) {
        ply->buffer[i] = tolower(ply->c);
        ply->c = fgetc(ply->fp);
        ply->column_number++;
        i++;
    }
    /* if it wasn't the word that ended, issue an error */
    if (!isspace(ply->c)) {
        if (i >= WORDSIZE) errorf(("Word too long at line %d, column %d", 
                    ply->line_number, ply->column_number));
        if (ply->c == EOF) errorf(("Unexpected end of file"));
        return 0;
    }
    ply->buffer[i] = '\0';
    return 1;
}

static int ply_read_chunk(p_ply ply, void *anybuffer, size_t size)
{
    char *buffer = (char *) anybuffer;
    size_t i = 0;
    assert(ply && ply->fp && ply->mode == PLY_READ);
    assert(ply->buffer_first <= ply->buffer_last);
    while (i < size) {
        if (ply->buffer_first < ply->buffer_last) {
            buffer[i] = ply->buffer[ply->buffer_first];
            ply->buffer_first++;
            i++;
        } else {
            ply->buffer_first = 0;
            ply->buffer_last = fread(ply->buffer, 1, BUFFERSIZE, ply->fp);
            if (ply->buffer_last <= 0) return 0;
        }
    }
    return 1;
}

static int ply_write_chunk(p_ply ply, void *anybuffer, size_t size)
{
    char *buffer = (char *) anybuffer;
    size_t i = 0;
    assert(ply && ply->fp && ply->mode == PLY_WRITE);
    assert(ply->buffer_last <= BUFFERSIZE);
    while (i < size) {
        if (ply->buffer_last < BUFFERSIZE) {
            ply->buffer[ply->buffer_last] = buffer[i];
            ply->buffer_last++;
            i++;
        } else {
            ply->buffer_last = 0;
            if (fwrite(ply->buffer, 1, BUFFERSIZE, ply->fp) < BUFFERSIZE)
                return 0;
        }
    }
    return 1;
}

static int ply_write_chunk_reverse(p_ply ply, void *anybuffer, size_t size)
{
    int ret = 0;
    ply_reverse(anybuffer, size);
    ret = ply_write_chunk(ply, anybuffer, size);
    ply_reverse(anybuffer, size);
    return ret;
}

static int ply_read_chunk_reverse(p_ply ply, void *anybuffer, size_t size)
{
    if (!ply_read_chunk(ply, anybuffer, size)) return 0;
    ply_reverse(anybuffer, size);
    return 1;
}

static void ply_reverse(void *anydata, size_t size)
{
    char *data = (char *) anydata;
    char temp;
    int i;
    for(i = 0; i < size/2; i++) {
        temp = data[i];
        data[i] = data[size-i-1];
        data[size-i-1] = temp;
    }
}

static void ply_init(p_ply ply)
{
    ply->c = ' ';
    ply->element = NULL;
    ply->nelements = 0;
    ply->line_number = 1;
    ply->column_number = 1;
    ply->idriver = NULL;
    ply->odriver = NULL;
    ply->buffer[0] = '\0';
    ply->buffer_first = ply->buffer_last = 0;
    ply->welement = 0;
    ply->wproperty = 0;
    ply->welement_index = 0;
    ply->wnentries = 0;
    ply->wentry_index = 0;
}

static void ply_element_init(p_ply_element element)
{
    element->name[0] = '\0';
    element->nelements = 0;
    element->property = NULL;
    element->nproperties = 0; 
}

static void ply_property_init(p_ply_property property)
{
    property->name[0] = '\0';
    property->type = -1;
    property->nentries_type = -1;
    property->value_type = -1;
    property->callback = (p_ply_cb) NULL;
    property->pdata = NULL;
    property->idata = 0;
}

static p_ply ply_alloc(void)
{
    p_ply ply = (p_ply) malloc(sizeof(t_ply));
    if (!ply) {
        errorf(("Out of memory"));
        return NULL;
    }
    ply_init(ply);
    return ply;
}

static p_ply_element ply_grow_elements(p_ply ply)
{
    p_ply_element element = NULL;
    int nelements;
    assert(ply); 
    element = ply->element;
    nelements = ply->nelements;
    assert(element || nelements == 0); 
    assert(!element || nelements > 0); 
    if (!element) 
        element = (p_ply_element) malloc(sizeof(t_ply_element));
    else element = (p_ply_element) 
        realloc(element, (nelements+1)*sizeof(t_ply_element));
    if (!element) {
        errorf(("Out of memory"));
        return 0;
    }
    ply->nelements = nelements+1;
    ply->element = element;
    ply_element_init(&element[nelements]);
    return &element[nelements];
}

static p_ply_property ply_grow_property(p_ply_element element)
{
    p_ply_property property = NULL;
    int nproperties;
    assert(element);
    if (!element) return NULL;
    property = element->property;
    nproperties = element->nproperties;
    assert(property || nproperties == 0);
    assert(!property || nproperties > 0);
    if (!property) property = (p_ply_property) malloc(sizeof(t_ply_property));
    else property = (p_ply_property)
        realloc(property, (nproperties+1)*sizeof(t_ply_property));
    if (!property) {
        errorf(("Out of memory"));
        return 0; 
    }
    element->nproperties = nproperties+1;
    element->property = property;
    ply_property_init(&property[nproperties]);
    return &property[nproperties];
}

static int ply_read_format(p_ply ply)
{
    assert(ply && ply->fp && ply->mode == PLY_READ);
    if (strcmp(ply->buffer, "format")) return 0;
    if (!ply_read_word(ply)) return 0;
    ply->format = ply_find_string(ply->buffer, ply_format_list);
    if (ply->format == -1) return 0;
    if (ply->format == PLY_ASCII) ply->idriver = &ply_idriver_ascii;
    else if (ply->format == ply_arch_endian()) 
        ply->idriver = &ply_idriver_binary;
    else ply->idriver = &ply_idriver_binary_reverse;
    if (!ply_read_word(ply)) return 0;
    if (strcmp(ply->buffer, "1.0")) return 0;
    if (!ply_read_word(ply)) return 0;
    return 1;
}

static int ply_read_comment(p_ply ply)
{
    assert(ply && ply->fp && ply->mode == PLY_READ);
    if (strcmp(ply->buffer, "comment")) return 0;
    while (ply->c != '\n' && ply->c != EOF)
        ply->c = fgetc(ply->fp);
    if (!ply_read_word(ply)) return 0;
    return 1;
}

static int ply_read_property(p_ply ply)
{
    p_ply_element element = NULL;
    p_ply_property property = NULL;
    /* make sure it is a property */
    if (strcmp(ply->buffer, "property")) return 0;
    element = &ply->element[ply->nelements-1];
    property = ply_grow_property(element);
    if (!property) return 0;
    /* get property type */
    if (!ply_read_word(ply)) return 0;
    property->type = ply_find_string(ply->buffer, ply_type_list);
    if (property->type < 0) return 0;
    if (property->type == PLY_LIST) {
        /* if it's a list, we need the base types */
        if (!ply_read_word(ply)) return 0;
        property->nentries_type = ply_find_string(ply->buffer, ply_type_list);
        if (property->nentries_type < 0) return 0;
        if (!ply_read_word(ply)) return 0;
        property->value_type = ply_find_string(ply->buffer, ply_type_list);
        if (property->value_type < 0) return 0;
    }
    /* get property name */
    if (!ply_read_word(ply)) return 0;
    strcpy(property->name, ply->buffer);
    if (!ply_read_word(ply)) return 0;
    return 1;
}

static int ply_read_element(p_ply ply)
{
    p_ply_element element = NULL;
    assert(ply && ply->fp && ply->mode == PLY_READ);
    if (strcmp(ply->buffer, "element")) return 0;
    /* allocate room for new element */
    element = ply_grow_elements(ply);
    if (!element) return 0;
    /* get element name */
    if (!ply_read_word(ply)) return 0;
    strcpy(element->name, ply->buffer);
    /* get number of elements of this type */
    if (!ply_read_word(ply)) return 0;
    if (sscanf(ply->buffer, "%ld", &element->nelements) != 1) {
        errorf(("Expected number at line %d, column %d", ply->line_number,
                    ply->column_number - strlen(ply->buffer)));
        return 0;
    }
    /* get all properties for this element (skipping comments) */
    if (!ply_read_word(ply)) return 0;
    while (ply_read_property(ply) || ply_read_comment(ply))
        /* do nothing */;
    return 1;
}

/* ----------------------------------------------------------------------
 * Error handling and debug
 * ---------------------------------------------------------------------- */
static void ply_print(p_ply ply)
{
    int i;
    (void) ply_print;
    printf("ply\nformat %s 1.0\n", ply_format_list[ply->format]);
    for (i = 0; i < ply->nelements; i++) {
        p_ply_element element = &ply->element[i];
        int j;
        printf("element %s %ld\n", element->name, element->nelements);
        for (j = 0; j < element->nproperties; j++) {
            p_ply_property property = &element->property[j];
            printf("property %s ", ply_type_list[property->type]);
            if (property->type == PLY_LIST)
                printf("%s %s ", ply_type_list[property->nentries_type],
                        ply_type_list[property->value_type]);
            printf("%s\n", property->name);
        }
    }
}

static void ply_error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
}

static int ply_arch_endian(void)
{
    unsigned long int i = 1;
    unsigned char *s = (unsigned char *) &i;
    if (*s == 1) return PLY_LITTLE;
    else return PLY_BIG;
}

static int ply_type_check(void)
{
    assert(sizeof(char) == 1);
    assert(sizeof(unsigned char) == 1);
    assert(sizeof(short) == 2);
    assert(sizeof(unsigned short) == 2);
    assert(sizeof(long) == 4);
    assert(sizeof(unsigned long) == 4);
    assert(sizeof(float) == 4);
    assert(sizeof(double) == 8);
    if (sizeof(char) != 1) return 0;
    if (sizeof(unsigned char) != 1) return 0;
    if (sizeof(short) != 2) return 0;
    if (sizeof(unsigned short) != 2) return 0;
    if (sizeof(long) != 4) return 0;
    if (sizeof(unsigned long) != 4) return 0;
    if (sizeof(float) != 4) return 0;
    if (sizeof(double) != 8) return 0;
    return 1;
}

/* ----------------------------------------------------------------------
 * Output handlers
 * ---------------------------------------------------------------------- */
static int oascii_int8(p_ply ply, double value)
{
    if (value > CHAR_MAX || value < CHAR_MIN) return 0;
    return fprintf(ply->fp, "%d ", (char) value) > 0;
}

static int oascii_uint8(p_ply ply, double value)
{
    if (value > UCHAR_MAX || value < 0) return 0;
    return fprintf(ply->fp, "%d ", (unsigned char) value) > 0;
}

static int oascii_int16(p_ply ply, double value)
{
    if (value > SHRT_MAX || value < SHRT_MIN) return 0;
    return fprintf(ply->fp, "%d ", (short) value) > 0;
}

static int oascii_uint16(p_ply ply, double value)
{
    if (value > USHRT_MAX || value < 0) return 0;
    return fprintf(ply->fp, "%d ", (unsigned short) value) > 0;
}

static int oascii_int32(p_ply ply, double value)
{
    if (value > LONG_MAX || value < LONG_MIN) return 0;
    return fprintf(ply->fp, "%d ", (int) value) > 0;
}

static int oascii_uint32(p_ply ply, double value)
{
    if (value > ULONG_MAX || value < 0) return 0;
    return fprintf(ply->fp, "%d ", (unsigned int) value) > 0;
}

static int oascii_float32(p_ply ply, double value)
{
    if (value < -FLT_MAX || value > FLT_MAX) return 0;
    return fprintf(ply->fp, "%g ", (float) value) > 0;
}

static int oascii_float64(p_ply ply, double value)
{
    if (value < -DBL_MAX || value > DBL_MAX) return 0;
    return fprintf(ply->fp, "%g ", value) > 0;
}

static int obinary_int8(p_ply ply, double value)
{
    char int8 = (char) value;
    if (value > CHAR_MAX || value < CHAR_MIN) return 0;
    return ply->odriver->ochunk(ply, &int8, sizeof(int8));
}

static int obinary_uint8(p_ply ply, double value)
{
    unsigned char uint8 = (unsigned char) value;
    if (value > UCHAR_MAX || value < 0) return 0;
    return ply->odriver->ochunk(ply, &uint8, sizeof(uint8)); 
}

static int obinary_int16(p_ply ply, double value)
{
    short int16 = (short) value;
    if (value > SHRT_MAX || value < SHRT_MIN) return 0;
    return ply->odriver->ochunk(ply, &int16, sizeof(int16));
}

static int obinary_uint16(p_ply ply, double value)
{
    unsigned short uint16 = (unsigned short) value;
    if (value > USHRT_MAX || value < 0) return 0;
    return ply->odriver->ochunk(ply, &uint16, sizeof(uint16)); 
}

static int obinary_int32(p_ply ply, double value)
{
    long int32 = (long) value;
    if (value > LONG_MAX || value < LONG_MIN) return 0;
    return ply->odriver->ochunk(ply, &int32, sizeof(int32));
}

static int obinary_uint32(p_ply ply, double value)
{
    unsigned long uint32 = (unsigned long) value;
    if (value > ULONG_MAX || value < 0) return 0;
    return ply->odriver->ochunk(ply, &uint32, sizeof(uint32));
}

static int obinary_float32(p_ply ply, double value)
{
    float float32 = (float) value;
    if (value > FLT_MAX || value < -FLT_MAX) return 0;
    return ply->odriver->ochunk(ply, &float32, sizeof(float32));
}

static int obinary_float64(p_ply ply, double value)
{
    return ply->odriver->ochunk(ply, &value, sizeof(value)); 
}

/* ----------------------------------------------------------------------
 * Input  handlers
 * ---------------------------------------------------------------------- */
static int iascii_int8(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value > CHAR_MAX || *value < CHAR_MIN) return 0;
    return 1;
}

static int iascii_uint8(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value > UCHAR_MAX || *value < 0) return 0;
    return 1;
}

static int iascii_int16(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value > SHRT_MAX || *value < SHRT_MIN) return 0;
    return 1;
}

static int iascii_uint16(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value > USHRT_MAX || *value < 0) return 0;
    return 1;
}

static int iascii_int32(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value > LONG_MAX || *value < LONG_MIN) return 0;
    return 1;
}

static int iascii_uint32(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtol(ply->buffer, &end, 10);
    if (*end || *value < 0) return 0;
    return 1;
}

static int iascii_float32(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtod(ply->buffer, &end);
    if (*end || *value < -FLT_MAX || *value > FLT_MAX) return 0;
    return 1;
}

static int iascii_float64(p_ply ply, double *value)
{
    char *end;
    if (!ply_read_word(ply)) return 0;
    *value = strtod(ply->buffer, &end);
    if (*end || *value < -DBL_MIN || *value > DBL_MAX) return 0;
    return 1;
}

static int ibinary_int8(p_ply ply, double *value)
{
    char int8;
    if (!ply->idriver->ichunk(ply, &int8, 1)) return 0;
    *value = int8;
    return 1;
}

static int ibinary_uint8(p_ply ply, double *value)
{
    unsigned char uint8;
    if (!ply->idriver->ichunk(ply, &uint8, 1)) return 0;
    *value = uint8;
    return 1;
}

static int ibinary_int16(p_ply ply, double *value)
{
    short int16;
    if (!ply->idriver->ichunk(ply, &int16, sizeof(int16))) return 0;
    *value = int16;
    return 1;
}

static int ibinary_uint16(p_ply ply, double *value)
{
    unsigned short uint16;
    if (!ply->idriver->ichunk(ply, &uint16, sizeof(uint16))) return 0;
    *value = uint16;
    return 1;
}

static int ibinary_int32(p_ply ply, double *value)
{
    long int32;
    if (!ply->idriver->ichunk(ply, &int32, sizeof(int32))) return 0;
    *value = int32;
    return 1;
}

static int ibinary_uint32(p_ply ply, double *value)
{
    unsigned long uint32;
    if (!ply->idriver->ichunk(ply, &uint32, sizeof(uint32))) return 0;
    *value = uint32;
    return 1;
}

static int ibinary_float32(p_ply ply, double *value)
{
    float float32;
    if (!ply->idriver->ichunk(ply, &float32, sizeof(float32))) return 0;
    *value = float32;
    ply_reverse(&float32, sizeof(float32));
    return 1;
}

static int ibinary_float64(p_ply ply, double *value)
{
    return ply->idriver->ichunk(ply, value, sizeof(double));
}

/* ----------------------------------------------------------------------
 * Constants
 * ---------------------------------------------------------------------- */
static t_ply_idriver ply_idriver_ascii = {
    {   iascii_int8, iascii_uint8, iascii_int16, iascii_uint16,
        iascii_int32, iascii_uint32, iascii_float32, iascii_float64,
        iascii_int8, iascii_uint8, iascii_int16, iascii_uint16,
        iascii_int32, iascii_uint32, iascii_float32, iascii_float64
    }, /* order matches e_ply_type enum */
    NULL,
    "ascii input"
};

static t_ply_idriver ply_idriver_binary = {
    {   ibinary_int8, ibinary_uint8, ibinary_int16, ibinary_uint16,
        ibinary_int32, ibinary_uint32, ibinary_float32, ibinary_float64,
        ibinary_int8, ibinary_uint8, ibinary_int16, ibinary_uint16,
        ibinary_int32, ibinary_uint32, ibinary_float32, ibinary_float64
    }, /* order matches e_ply_type enum */
    ply_read_chunk,
    "binary input"
};

static t_ply_idriver ply_idriver_binary_reverse = {
    {   ibinary_int8, ibinary_uint8, ibinary_int16, ibinary_uint16,
        ibinary_int32, ibinary_uint32, ibinary_float32, ibinary_float64,
        ibinary_int8, ibinary_uint8, ibinary_int16, ibinary_uint16,
        ibinary_int32, ibinary_uint32, ibinary_float32, ibinary_float64
    }, /* order matches e_ply_type enum */
    ply_read_chunk_reverse,
    "reverse binary input"
};

static t_ply_odriver ply_odriver_ascii = {
    {   oascii_int8, oascii_uint8, oascii_int16, oascii_uint16,
        oascii_int32, oascii_uint32, oascii_float32, oascii_float64,
        oascii_int8, oascii_uint8, oascii_int16, oascii_uint16,
        oascii_int32, oascii_uint32, oascii_float32, oascii_float64
    }, /* order matches e_ply_type enum */
    NULL,
    "ascii output"
};

static t_ply_odriver ply_odriver_binary = {
    {   obinary_int8, obinary_uint8, obinary_int16, obinary_uint16,
        obinary_int32, obinary_uint32, obinary_float32, obinary_float64,
        obinary_int8, obinary_uint8, obinary_int16, obinary_uint16,
        obinary_int32, obinary_uint32, obinary_float32, obinary_float64
    }, /* order matches e_ply_type enum */
    ply_write_chunk,
    "binary output"
};

static t_ply_odriver ply_odriver_binary_reverse = {
    {   obinary_int8, obinary_uint8, obinary_int16, obinary_uint16,
        obinary_int32, obinary_uint32, obinary_float32, obinary_float64,
        obinary_int8, obinary_uint8, obinary_int16, obinary_uint16,
        obinary_int32, obinary_uint32, obinary_float32, obinary_float64
    }, /* order matches e_ply_type enum */
    ply_write_chunk_reverse,
    "reverse binary output"
};
