#include <cstdio>
#include <cstdlib>
#include <algorithm>

#include "glutwm.h"
#include "soupwindow.h"

SoupWindow::
~SoupWindow()
{
    ;
}

void
SoupWindow::
Init(int id)
{
    Window::Init(id);
    glShadeModel(GL_FLAT);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    glDisable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_COLOR_MATERIAL);
    GLfloat s[] = {.3, .3, .3, 1};
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, s);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    GLfloat a[] = {0, 0, 0, 1};
    float l[] = {0.75, 0.75, 0.75, 0};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, a);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, l);
    glLightfv(GL_LIGHT0, GL_SPECULAR, l);
    glLightfv(GL_LIGHT0, GL_AMBIENT, a);
}


void
SoupWindow::
DrawSoup(void)
{
    switch (mode) {
        case MODE_TRIANGLES:
            glEnable(GL_LIGHTING);
            glBegin(GL_TRIANGLES); 
            for (int i = 0; i < soup->triangle.GetSize(); i++) {
                const Soup::Triangle &t = soup->triangle[i];
                const Vector3 &v0 = soup->vertex[t.index[0]];
                const Vector3 &v1 = soup->vertex[t.index[1]];
                const Vector3 &v2 = soup->vertex[t.index[2]];
                const Vector3 &n = Cross(v1-v0, v2-v0);
                glNormal3d(n[X], n[Y], n[Z]);
                glVertex3d(v0[X], v0[Y], v0[Z]);
                glVertex3d(v1[X], v1[Y], v1[Z]);
                glVertex3d(v2[X], v2[Y], v2[Z]);
            }
            glEnd();
            break;
        case MODE_POINTS:
            glDisable(GL_LIGHTING);
            glPointSize(1);
            glBegin(GL_POINTS);
            for (int i = 0; i < soup->vertex.GetSize(); i++) {
                const Vector3 &v = soup->vertex[i];
                glVertex3d(v[X], v[Y], v[Z]);
            }
            glEnd();
            break;
    }
}

void
SoupWindow::
SetLightPosition(float *f)
{
    for (int i = 0; i < 4; i++) light[i] = f[i];
}

void 
SoupWindow::Display(void)
{
    int w = GetWidth(), h = GetHeight();
    int x0 = GetX0(), y0 = GetY0();
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (!soup) return;
    glViewport(x0, y0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    double a = zoom.Scale();
    glOrtho(-size/2, size/2, -size/2*h/w, size/2*h/w, -size/2*a, size/2*a);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glLightfv(GL_LIGHT0, GL_POSITION, light);
    double x, y, z;
    trans.Translation(&x, &y, &z);
    glTranslated(size*x, y*size*h/w, z);
    a = ball.Rotation(&x, &y, &z);
    glRotated(a, x, y, z);
    a = zoom.Scale();
    glScaled(a, a, a);
    glTranslated(-center[X], -center[Y], -center[Z]);
    glColor3d(1, 1, 1);
	DrawSoup();
}

void
SoupWindow::
SetSoup(Soup *_soup)
{
    soup = _soup;
}

Soup * 
SoupWindow::
GetSoup(void)
{
    return soup;
}

// estimate center and bounding box using medians to make it robust
int
SoupWindow::
Fit(void)
{
    if (!soup) {
        errorf(("No buffer setup"));
        return 0;
    }
    // solve for x, y and z independently
    Array<double> a, b, c;
    int nvertex = soup->vertex.GetSize();
    if (nvertex <= 0) {
        center = Vector3(0, 0, 0);
        size = 1;
        return 0;
    }
    if (!a.Resize(nvertex) || !b.Resize(nvertex) || !c.Resize(nvertex)) {
        errorf(("Unable to allocate vectors"));
        return 0;
    }
    // copy coordinates to temp buffers
    for (int i = 0; i < nvertex; i++) {
        const Vector3 &v = soup->vertex[i];
        a[i] = v[X];
        b[i] = v[Y];
        c[i] = v[Z];
    }
    // center is median independently in  3 dimensions
    std::nth_element(&a[0], &a[nvertex/2], &a[nvertex]);
    std::nth_element(&b[0], &b[nvertex/2], &b[nvertex]);
    std::nth_element(&c[0], &c[nvertex/2], &c[nvertex]);
    center[X] = a[nvertex/2];
    center[Y] = b[nvertex/2];
    center[Z] = c[nvertex/2];
    // get median distance to median center
    for (int i = 0; i < nvertex; i++) {
        a[i] = (soup->vertex[i] - center).Norm();
    }
    b = a;
    std::nth_element(&b[0], &b[nvertex/2], &b[nvertex]);
    double distance = b[nvertex/2];
    // compute median deviation from median distance to median center
    for (int i = 0; i < nvertex; i++) {
        b[i] = fabs(a[i] - distance);
    }
    c = b;
    std::nth_element(&c[0], &c[nvertex/2], &c[nvertex]);
    double deviation = c[nvertex/2];
    // discard points that are too far away from center
    BBox3 box;
    for (int i = 0; i < nvertex; i++) {
        if (b[i] < 2*deviation)
            box.Merge(soup->vertex[i]);
    }
    Vector3 s = box.GetSize();
    size = s[X] + s[Y] + s[Z];
    debugf(("Soup center: (%f, %f, %f)", center[X], center[Y], center[Z]));
    debugf(("Soup size: %f", size));
    return 1;
}

void
SoupWindow::
Button(int button, int state, int x, int y)
{
    switch (button) {
        case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN) ball.Down(x, y);
            else ball.Up(x, y);
            break;
        case GLUT_MIDDLE_BUTTON:
            if (state == GLUT_DOWN) trans.Down(x, y);
            else trans.Up(x, y);
            break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN) zoom.Down(x, y);
            else zoom.Up(x, y);
            break;
    }
    GlutWMPostRedisplay();
}

void
SoupWindow::
Keyboard(unsigned char c, int x, int y)
{
    switch (c)  {
        case 'f':
            // reset viewing transformation
            ball.Set(); zoom.Set(); trans.Set();
            Fit();
            GlutWMPostRedisplay();
            break;
        case 'm':
            if (mode == MODE_POINTS) mode = MODE_TRIANGLES;
            else if (mode == MODE_TRIANGLES) mode = MODE_POINTS;
            GlutWMPostRedisplay();
            break;
        default:
            Window::Keyboard(c, x, y);
            break;
    }
}

void
SoupWindow::Motion(int x, int y)
{
    ball.Motion(x, y);
    zoom.Motion(x, y);
    trans.Motion(x, y);
    GlutWMPostRedisplay();
}

#ifndef MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#endif
#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

void
SoupWindow::
Reshape(int w, int h, int x0, int y0)
{
    Window::Reshape(w, h, x0, y0);
    ball.Reset(w/2, h/2, MIN(w/2, h/2));
    zoom.Reset(w, h, x0, y0);
    trans.Reset(w, h, x0, y0); 
}

SoupWindow::SoupWindow(void) 
{
    soup = NULL;
    list = 0;
    light[0] = 0;
    light[1] = 0;
    light[2] = 1;
    light[3] = 0;
    mode = MODE_TRIANGLES;
}
